#!/usr/bin/env python3
"""Routines related to clustering."""
import logging
import time
from functools import wraps

import numpy as np
import pandas as pd
from hdbscan import HDBSCAN, RobustSingleLinkage
from sklearn import metrics
from sklearn.cluster import DBSCAN, OPTICS

from .config_parser import UndefinedConfigValue
from .domains import Domain
from .metrics import calc_distance_matrix
from .outlier_removal import filter_outliers

logger = logging.getLogger(__name__)


def sort_df_by_cluster_size(df):
    """Return a version of df with data sorted by cluster_size.

    Sort df so that clusters with more members are put at the top of the
    dataframe. The exception if the "-1" label, which, if present, will
    always remain at the top. Handy if results are to be plotted.

    Args:
        df (pandas.DataFrame): Dataframe with clustering info.

    Returns:
        pandas.DataFrame: Copy of input df sorted by cluster size.

    """
    # The labels may have been reassigned if the df was passed through
    # an outlier removal routine. Let's keep track of the original ones.
    # This will help keeping obs removed from a cluster by the outlier
    # removed routine close to their origin cluster.
    original_cluster_label_col = "cluster_label"
    if "original_cluster_label" in df.columns:
        original_cluster_label_col = "original_cluster_label"

    # Temporarily add a column with cluster sizes for sorting
    unique_labels, member_counts = np.unique(
        df[original_cluster_label_col], return_counts=True
    )
    label2count_dict = dict(zip(unique_labels, member_counts))
    df["parent_cluster_size"] = [
        label2count_dict[label] for label in df[original_cluster_label_col]
    ]
    # Sort dataframe according to frequency, but keep -1 at the top
    df = pd.concat(
        [
            df[df["cluster_label"] == -1],
            df[df["cluster_label"] == -3],
            df[df["cluster_label"] == -4],
            df[~df["cluster_label"].isin([-1, -3, -4])].sort_values(
                [
                    "parent_cluster_size",
                    original_cluster_label_col,
                    "cluster_label",
                ],
                ascending=[False, True, False],
            ),
        ]
    )

    # Reassign labels so that fewer members leads to larger labels.
    # Note that the df unique method does not sort, so this method won't
    # mess up the sorting performed above.
    unique_labels = df[original_cluster_label_col].unique()
    _labels_old2new = {
        old: new
        for new, old in enumerate(lab for lab in unique_labels if lab >= 0)
    }

    @np.vectorize
    def labels_old2new(old):
        if old < 0:
            return old
        return _labels_old2new[old]

    for col in ["cluster_label", "original_cluster_label"]:
        try:
            df[col] = labels_old2new(df[col])
        except KeyError:
            pass

    return df.drop("parent_cluster_size", axis=1).reset_index(drop=True)


def weights_dict_to_np_array(
    df, pairwise_diff_weights=None, skip=("id", "time_utc"), default=1
):
    """Convert pairwise_diff_weights into a numpy array.

    Takes a pandas dataframe and a {column_name:weight} dictionary and returns
    an array of weights to be passed to the routine that calculates the
    distance matrix.

    Columns "lat" and "lon" in df are treated specially, in that they are
    not assigned a weight individually, but rather a single weight gets
    assigned to the "geo_dist" property.

    Args:
        df (pandas.Dataframe): Dataframe with observations.
        pairwise_diff_weights (dict):  {df_column_name:weight} dictionary.
            Default value = None.
        skip: df columns that will not enter the clustering and should
            therefore be skipped. Default value = ("id", "time_utc")
        default: Default weight to be assigned for a non-skipped df column if
            the column name is present in df but not in pairwise_diff_weights.
            Default value = 1.

    Returns:
        numpy.ndarray: weights to be passed to the routine that calculates the
            distance matrix.

    Raises:
        ValueError: If the dataframe 'lat' column is not followed by the 'lon'
            column.

    """
    if df.columns.get_loc("lon") - df.columns.get_loc("lat") != 1:
        raise ValueError("'lat' column is not followed by 'lon' column")

    weights = []
    col2weight = {c: ("geo_dist" if c == "lon" else c) for c in df.columns}
    for col in df.columns[~df.columns.isin(list(skip) + ["lat"])]:
        try:
            weights.append(pairwise_diff_weights[col2weight[col]])
        except (KeyError, TypeError):
            weights.append(default)
    return np.array(weights, dtype=np.float64)


def get_silhouette_samples(df, distance_matrix):
    """Calculate silhouette scores for every obs in df.

    Args:
        df (pandas.Dataframe): Dataframe with observations.
        distance_matrix (HollowSymmetricMatrix): Obs distance matrix.

    Returns:
        pandas.DataFrame: Copy of df with added silhouette samples info.

    """
    # We will only consider true clusters for the calculation of
    # the silhouette coeff, i.e., points with label<0 will be excluded from
    # the calculation. This means that we need to use a slice of the distance
    # matrix in the calculation corresponding to the valid points.
    i_valid_obs = df.reset_index(drop=True).index[df["cluster_label"] >= 0]
    if len(i_valid_obs) == 0:
        logger.warning("No valid obs to calculate silhouette coeffs with")
        return -np.ones(len(df.index))

    reduced_labels = df["cluster_label"].iloc[i_valid_obs]
    if len(reduced_labels.unique()) == 1:
        logger.warning("Only one cluster to calculate silhouette coeffs with")
        return np.ones(len(df.index))

    # Using an intermediate np array is about 10x faster than
    # using df.at[iloc, 'silhouette_score'] = coeff
    all_silhouette_scores = np.empty(len(df.index))
    all_silhouette_scores[:] = np.nan

    reduced_silhouette_scores = metrics.silhouette_samples(
        X=distance_matrix.subspace(i_valid_obs),
        labels=reduced_labels,
        metric="precomputed",
    )
    for iloc, coeff in zip(i_valid_obs, reduced_silhouette_scores):
        all_silhouette_scores[iloc] = coeff

    return all_silhouette_scores


def run_clustering_on_df(
    df,
    method="hdbscan",
    distance_matrix=None,
    distance_matrix_optimize_mode="memory",
    skip=("id", "time_utc"),
    weights_dict=None,
    eps=15,  # eps applies only to dbscan
    min_cluster_size=3,  # min_cluster_size applies only to hdbscan
    min_samples=3,
    n_jobs=-1,
    outlier_rm_method=None,
    max_num_refine_iter=50,
    max_n_stdev_around_mean=2.0,
    trunc_perc=0.25,
    calc_silhouette_samples=True,
):
    """Low-level clustering routine.

    Args:
        df (pandas.Dataframe): Dataframe with observations.
        method (str): {"hdbscan", "dbscan", "rsl", "optics"}
            Clustering method.
        distance_matrix (HollowSymmetricMatrix): Obs distance matrix.
            Default value = None.
        distance_matrix_optimize_mode (str): The distance matrix optimization
            mode. Default value = "memory".
        skip (tuple): Columns to skip in df.
            Default value = ("id", "time_utc").
        weights_dict (dict): obs_name --> obs_weight map. Default value = None.
        eps (float):  DBSCAN's eps parameter. Default value = 15.0.
        min_cluster_size (int):  HDBSCAN's min_cluster_size. Default value = 3.
        min_samples (int): (H)DBSCAN's min_samples. Default value = 3.
        n_jobs (int): Max number of local-host parallel jobs.
            Default value = -1.
        outlier_rm_method (str):  Outlier removal method. Default value = None.
        max_num_refine_iter (int):  Max number of iterations for the
            "iterative" ourlier removal method. Default value = 50.
        max_n_stdev_around_mean (float):  Max number of stdev from cluster obs
            mean for an observation to be considered OK in the "iteractive"
            outlier removal method. Default value = 2.0.
        trunc_perc (float):  Percentage used in array truncation when
            calculating stdevs and means in the "iterative" outlier removal
            method. Default value = 0.25.
        calc_silhouette_samples (bool):  Calculate or not silhouette_samples?
            Default value = True.

    Returns:
        pandas.DataFrame: Copy of df with clustering added info.

    Raises:
        NotImplementedError: If the `method` choice is not valid.

    """
    method = method.lower()
    # Compute clustering using DBSCAN or HDBSCAN
    if method not in ["dbscan", "hdbscan", "rsl", "optics"]:
        raise NotImplementedError('Method "{}" not available.'.format(method))
    if len(df.index) == 0:
        logger.warning("Dataframe has no rows")
        return df

    # Set weights to be used in the metrics for the various
    # generalised distances. The distances used in the metrics
    # will be used_dist(i) = weight(i)*real_dist(i)
    weights = weights_dict_to_np_array(df, weights_dict)

    # We will not do any df = StandardScaler().fit_transform(df),
    # as we'll use a metric based on earth distances

    if distance_matrix is None:
        # My tests indicate that passing a pre-computed distance matrix to
        # dbscan can be up to 2.5x faster than passing a metrics function (if
        # they both are written in pure python) to fit df. If they are both
        # written in fortran and interfaced via f2py3, or then written in
        # python but jit-compiled with numba, then the relative speed up
        # can reach up to 120x.
        distance_matrix = calc_distance_matrix(
            # Drop columns that won't be used in the clustering
            df.drop(list(skip), axis=1),
            weights,
            optimize_mode=distance_matrix_optimize_mode,
            num_threads=n_jobs,
        )

    # Running clustering with the computed distance matrix
    logger.debug("    > Running clustering method '%s'...", method)
    tstart = time.time()
    if method == "dbscan":
        db = DBSCAN(
            eps=eps,
            min_samples=min_samples,
            metric="precomputed",
            n_jobs=n_jobs,
        ).fit(distance_matrix)
        core_samples_mask = np.zeros_like(db.labels_, dtype=bool)
        core_samples_mask[db.core_sample_indices_] = True
        df["is_core_point"] = core_samples_mask
    elif method == "hdbscan":
        # For more info on the parameters, see
        # <https://hdbscan.readthedocs.io/en/latest/parameter_selection.html>
        db = HDBSCAN(
            min_samples=min_samples,
            min_cluster_size=min_cluster_size,
            metric="precomputed",
            core_dist_n_jobs=n_jobs,
            allow_single_cluster=True,
            # Default cluster_selection_method: 'eom'. Sometimes it leads to
            # clusters that are too big. Using 'leaf' seems better.
            cluster_selection_method="leaf",
        ).fit(distance_matrix)
    elif method == "optics":
        db = OPTICS(
            min_samples=min_samples,
            min_cluster_size=min_cluster_size,
            n_jobs=n_jobs,
            metric="precomputed",
        ).fit(distance_matrix)
    elif method == "rsl":
        db = RobustSingleLinkage(
            # cut: The reachability distance value to cut the cluster
            #      heirarchy at to derive a flat cluster labelling.
            cut=eps,  # default=0.4
            # Reachability distances will be computed with regard to the
            # k nearest neighbors.
            k=min_samples,  # default=5
            # Ignore any clusters in the flat clustering with size less
            # than gamma, and declare points in such clusters as noise points.
            gamma=min_cluster_size,  # default=5
            metric="precomputed",
        ).fit(distance_matrix)
    logger.debug(
        "      * Done with %s. Elapsed: %.2fs", method, time.time() - tstart,
    )
    # Update df with cluster label info. It is important that this is done
    # right before calling filter_outliers, as the filter_outliers function
    # expects 'cluster_label' to be the last column in the dataframe
    df["cluster_label"] = db.labels_

    # Refine clustering if requested
    # It is important to have 'cluster_label' as the last column
    # when running the iterative refine routine
    if outlier_rm_method:
        df = filter_outliers(
            df,
            db=db,
            outlier_rm_method=outlier_rm_method,
            # Args that apply only to LOF
            distance_matrix=distance_matrix,
            # Args that only apply for the iterative method
            skip=skip,
            max_num_refine_iter=max_num_refine_iter,
            max_n_stdev_around_mean=max_n_stdev_around_mean,
            trunc_perc=trunc_perc,
            weights=weights,
            method=method,
            weights_dict=weights_dict,
            eps=eps,
            min_cluster_size=min_cluster_size,
            min_samples=min_samples,
            n_jobs=n_jobs,
            reclustering_function=self_consistent_reclustering,
        )

    # Calculate silhouette scores and update df with this.
    # We'll be reordering the dataframe later, and calculating the score
    # before doing that avoids the need to also reorder the dist matrix.
    if calc_silhouette_samples:
        df["silhouette_score"] = get_silhouette_samples(df, distance_matrix)

    return df


# self_consistent_reclustering is used in the "reclustering" outlier removal
# method. Disable logging in the function to reduce output clutter.
def suspendlogging(func):
    """Suspend logging in the function it decorates.

    Args:
        func (function): The function to be decorated.

    Returns:
        function: The decorated function.

    """
    # Adapted from: <https://stackoverflow.com/questions/7341064/
    #               disable-logging-per-method-function>

    @wraps(func)
    def _inner(*args, **kwargs):
        previousloglevel = logger.getEffectiveLevel()
        try:
            logger.setLevel(logging.WARN)
            return func(*args, **kwargs)
        finally:
            logger.setLevel(previousloglevel)

    return _inner


@suspendlogging
def self_consistent_reclustering(df, distance_matrix, **kwargs):
    """Recluster obs in df until further clustering on df returns df itself.

    Args:
        df (pandas.Dataframe): Pandas dataframe with observations and initial
            clustering info.
        distance_matrix (HollowSymmetricMatrix): Matrix of distances between
            the obs in df.
        **kwargs: Keyword args passed to run_clustering_on_df.

    Returns:
        pandas.Dataframe: Copy of df where obs have been clustered multiple
            times until run_clustering_on_df(df, ...) = df.

    """
    n_iter = 0
    n_removed = 0
    n_noise = np.count_nonzero(df["cluster_label"] < 0)
    while n_noise != 0:
        noise_mask = df["cluster_label"] < 0
        df_noise = df[noise_mask]
        i_valid_obs = np.nonzero((~noise_mask).values)[0]

        df = df.drop(df[noise_mask].index)
        df = df.drop(["cluster_label"], axis=1)
        df_rec = run_clustering_on_df(
            df,
            outlier_rm_method=None,
            distance_matrix=distance_matrix.subspace(i_valid_obs),
            calc_silhouette_samples="silhouette_score" in df.columns,
            **kwargs,
        )
        n_noise = np.count_nonzero(df_rec["cluster_label"] < 0)
        df = pd.concat([df_noise, df_rec]).sort_index()
        n_iter += 1
        n_removed += n_noise
        logger.debug("        * Iter #%d: %d new noise pts", n_iter, n_noise)
    logger.debug("    > Done with reclustering. %d removed obs.", n_removed)
    return df


def _cluster_netatmo_obs_one_domain(df, config, **kwargs):
    """Cluster netatmo obs inside a given domain.

    Helper for the main routine cluster_netatmo_obs

    Args:
        df (pandas.Dataframe): Pandas dataframe with observations.
        config (ParsedConfig): Parsed configs.
        **kwargs: Keyword args passed to run_clustering_on_df.

    Returns:
        pandas.DataFrame: Copy of df with clustering added info.

    """
    time_start_clustering = time.time()
    logger.debug("Performing clustering...")
    outlier_rm_method = config.get_clustering_opt("outlier_removal.method")
    df = run_clustering_on_df(
        df=df,
        method=config.general.clustering_method,
        distance_matrix_optimize_mode=config.general.custom_metrics_optimize_mode,
        weights_dict=config.get_clustering_opt("obs_weights"),
        eps=config.get_clustering_opt("eps"),
        min_cluster_size=config.get_clustering_opt("min_cluster_size"),
        min_samples=config.get_clustering_opt("min_samples"),
        outlier_rm_method=outlier_rm_method,
        max_num_refine_iter=config.get_clustering_opt(
            "outlier_removal.{}.max_n_iter".format(outlier_rm_method)
        ),
        max_n_stdev_around_mean=config.get_clustering_opt(
            "outlier_removal.{}.max_n_stdev".format(outlier_rm_method)
        ),
        **kwargs,
    )
    time_end_clustering = time.time()
    logger.debug(
        "Done with clustering. Elapsed: %.2fs",
        time_end_clustering - time_start_clustering,
    )

    return df


def cluster_netatmo_obs(df, config, **kwargs):
    """Cluster NetAtmo observations.

    Args:
        df (pandas.Dataframe): Pandas dataframe with observations.
        config (ParsedConfig): Parsed configs.
        **kwargs: Keyword args passed to _cluster_netatmo_obs_one_domain.

    Returns:
        pandas.DataFrame: Copy of df with clustering added info.

    """
    # Load domain configs from config
    domain = Domain.construct_from_dict(config.domain)

    # Trim obs to keep only those inside domain. Do this here instead of
    # upon reading so we do this only once.
    n_rmvd = len(df.index)
    df = domain.trim_obs(df)
    n_rmvd -= len(df.index)
    if n_rmvd > 0:
        logger.debug("Removed %d obs outside %s domain", n_rmvd, domain.name)

    df_rejected = None
    if domain.n_subdomains > 1:
        # Perform preliminary clustering on small domains in order to
        # eliminate stations at low cpu/mem cost (O(n^2))
        logger.debug(
            "DTG=%s: Preliminary clustering over %d subdomains...",
            df.metadata_dict["dtg"],
            domain.n_subdomains,
        )
        domain_split_dfs = []
        # Silhouette samples are not needed at this point
        pre_clustering_kwargs = kwargs.copy()
        pre_clustering_kwargs["calc_silhouette_samples"] = False
        for isubd, subdomain in enumerate(domain.split()):
            df_sub = subdomain.trim_obs(df)

            min_cluster_size = config.get_clustering_opt("min_cluster_size")
            if min_cluster_size is UndefinedConfigValue:
                min_cluster_size = config.get_clustering_opt("min_samples")
            if len(df_sub.index) < min_cluster_size:
                logger.debug(
                    "DTG=%s: Too few obs (=%d) in subdomain %d/%d. Rejecting.",
                    df.metadata_dict["dtg"],
                    len(df_sub.index),
                    isubd + 1,
                    domain.n_subdomains,
                )
                df_sub["cluster_label"] = -1
                domain_split_dfs.append(df_sub)
                continue
            logging.debug(
                "DTG=%s: Pre-clustering %d obs in subdomain %d/%d",
                df.metadata_dict["dtg"],
                len(df_sub.index),
                isubd + 1,
                domain.n_subdomains,
            )
            df_sub = _cluster_netatmo_obs_one_domain(
                df=df_sub, config=config, **pre_clustering_kwargs
            )

            domain_split_dfs.append(df_sub)

        # Gather results from preliminary clustering in a single dataframe
        df_rejoined_split = pd.concat(domain_split_dfs, ignore_index=True)

        # Mark with flag -3 obs that were removed by outlier removal methods at
        # the pre-clustering stage (which are flagged with "-2")
        df_rejoined_split["cluster_label"].mask(
            df_rejoined_split["cluster_label"] == -2, -3, inplace=True
        )

        # Check for stations missed in the splits, (unlikely, but just in case)
        df_split_missed = df[~df["id"].isin(df_rejoined_split["id"])]
        if len(df_split_missed.index) > 0:
            logger.warning(
                "%d obs missed during subdomain split. Rejecting.",
                len(df_split_missed.index),
            )
            # Mark missed obs with cluster label -4
            df_split_missed["cluster_label"] = -4
            df_rejoined_split = pd.concat(
                [df_rejoined_split, df_split_missed], ignore_index=True
            )

        # Reset df: Only accepted obs will be passed on to the whole-domain
        # clustering. Rejections will be added again to df after that.
        df_rejected = df_rejoined_split[
            df_rejoined_split["cluster_label"] < 0
        ].copy()
        cols_to_drop = [c for c in df_rejected.columns if c not in df.columns]
        df = df_rejoined_split
        df = df[~df["id"].isin(df_rejected["id"])].drop(cols_to_drop, axis=1)
        df = df.reset_index(drop=True)

    # Run whole-domain clustering, regardless of splitting in subdomains or not
    if domain.n_subdomains > 1:
        logger.debug(
            "DTG=%s: Main clustering over whole domain...",
            df.metadata_dict["dtg"],
        )
    df = _cluster_netatmo_obs_one_domain(df=df, config=config, **kwargs)

    if df_rejected is not None:
        # Put back eventual obs rejected at the pre-clustering step
        if "original_cluster_label" in df.columns:
            df_rejected["original_cluster_label"] = df_rejected[
                "cluster_label"
            ].copy()
        df = pd.concat([df, df_rejected], ignore_index=True)

    # Now we're done.
    return df


def report_clustering_results(df):
    """Print some relevant info about the clustering.

    Args:
        df (pandas.DataFrame): Dataframe with clustering info.

    """
    if len(df.index) == 0:
        logger.info("Dataframe has no rows. No results to report.")
        return

    n_obs = len(df.index)
    noise_data_df = df[df["cluster_label"] < 0]
    n_noise_clusters = len(noise_data_df["cluster_label"].unique())
    noise_count = len(noise_data_df)
    n_clusters = len(df["cluster_label"].unique()) - n_noise_clusters
    n_accepted = n_obs - noise_count

    logger.info("Number of obs passed to the clustering routine: %d", n_obs)
    logger.info("Estimated number of clusters: %d", n_clusters)
    logger.info("Estimated number of accepted obs: %d", n_accepted)
    logger.info(
        "Estimated number of noise obs: %d (%.1f%% rejection rate)",
        noise_count,
        100.0 * noise_count / n_obs,
    )
    try:
        silhouette_score = df["silhouette_score"].mean(skipna=True)
        logger.info("Mean silhouette score: %.3f", silhouette_score)
    except KeyError:
        pass



#!/usr/bin/env python3
"""Calculation of custom metrics used in the code."""
import atexit
import logging
import time

import numba
import numpy as np
from numba import njit, prange

from .hollow_symmetric_matrix import (
    HollowSymmetricMatrix,
    _data_index_to_matrix_index,
)

logger = logging.getLogger(__name__)


@njit(cache=True)
def haversine_distance(point1, point2):
    """Return the Haversine distance between point1 and point2."""
    lat1 = np.deg2rad(point1[0])
    lat2 = np.deg2rad(point2[0])
    dlat = lat2 - lat1
    dlon = np.deg2rad(point2[1] - point1[1])
    param_a = (
        np.sin(dlat * 0.5) ** 2
        + np.cos(lat1) * np.cos(lat2) * np.sin(dlon * 0.5) ** 2
    )
    # earth_mean_diameter = 12742.0176 Km
    rtn = 12742.0176 * np.arcsin(np.sqrt(param_a))
    return rtn


@njit(cache=True)
def get_obs_norm_factors(obs_values):
    """Return multiplicative factor to normalise obs_values."""
    stdev = np.std(obs_values)
    obs_mean = np.mean(obs_values)
    if (
        (abs(obs_mean) > 1e-5 and abs(stdev / obs_mean) < 1e-5)
        or abs(obs_mean) <= 1e-5
        and stdev < 1e-7
    ):  # pylint: disable=chained-comparison
        rtn = 0.0
    else:
        rtn = 1.0 / stdev
    return rtn


@njit("f4[:](f8[:, :], f8[:])", parallel=True, cache=True)
def calc_distance_matrix_numba(df, weights):
    """Calculate distance matrix using python+numba.

    Args:
        df (numpy.ndarray): Multidimensional numpy array containing the data
            entries, obtained from a pandas dataframe (numba doesn't work with
            pandas datafames).
        weights (numpy.array): Weights chosen for each observation parameter.
            The weigts determine the relative importance of the observation
            parameters w.r.t. each other.

    Returns:
        numpy.ndarray: Distance matrix.

    """
    nrows, ncols = df.shape

    # Get normalisation factors so that observations in different
    # scales have comparable diff values
    # Remember: The first entry in weights refers to geo_dist and
    #           the first two columns of df are (lat, lon)
    obs_norm_factors = np.ones(ncols - 1, dtype=np.float32)
    for i in range(1, len(obs_norm_factors)):
        obs_norm_factors[i] = get_obs_norm_factors(df[:, i + 1])

    # weights can be used by the users to give extra or
    # lower weight to a given observation type. The weights
    # are applied to the normalised values of the obd diffs
    #
    # We are not normalising geodists. Therefore, the normalised obs
    # dists should have the same units as the geodists (km).
    # Let's assign 10 km/normalized_obs_unit_diff by default
    # This unit conversion is highly arbitrary...

    weights_internal = weights
    # Set any negative weight value to zero
    weights_internal = np.where(weights_internal < 0, 0.0, weights_internal)
    weights_internal *= obs_norm_factors

    n_dists = (nrows * (nrows - 1)) // 2
    rtn = np.zeros(n_dists, dtype=np.float32)
    # For better parallel performance, loop over single idist instead of doing
    # a nested "for i in prange(nrows): for j in range(i+1, nrows):".
    # If one wishes to reverse this, then use func _matrix_index_to_data_index
    # or note that idist = i*nrows - i*(i+1)//2 + (j - i - 1)
    # NB.: Initialising i and j as integers, otherwise numba v0.50.1 fails
    # silently and gives wrong results in python 3.8 with "parallel=True".
    # Strangely, the results are correct without initialisation when using
    # python 3.6 or when setting parallel=False.
    i, j = np.zeros(2, dtype=np.int64)
    for idist in prange(n_dists):  # pylint: disable=not-an-iterable
        i, j = _data_index_to_matrix_index(nrows, idist, check_bounds=False)
        rtn[idist] = weights_internal[0] * haversine_distance(
            df[i], df[j]
        ) + np.sum(
            np.sqrt((weights_internal[1:] * (df[j, 2:] - df[i, 2:])) ** 2)
        )

    return rtn


def calc_distance_matrix(df, weights, optimize_mode, num_threads=-1):
    """Calculate distance matrix between obs in dataframe df.

    Args:
        df (pandas.Dataframe): Input data.
        weights (numpy.array): Weights chosen for each observation parameter.
            The weigts determine the relative importance of the observation
            parameters w.r.t. each other.
        optimize_mode: How the distance matrix is to be calculated and stored.
            This is passed onto the constructor for the class
            netatmoqc.hollow_symmetric_matrix.HollowSymmetricMatrix.
        num_threads: Max number of threads used for the computation.
            (Default value = -1)

    Returns:
        netatmoqc.hollow_symmetric_matrix.HollowSymmetricMatrix: The distance
            matrix.

    """
    logger.debug("    > Calculating distance matrix...")
    tstart = time.time()

    if num_threads > 0:
        original_nthreads = numba.get_num_threads()
        numba.set_num_threads(num_threads)
        atexit.register(numba.set_num_threads, original_nthreads)

    rtn = HollowSymmetricMatrix(
        data=calc_distance_matrix_numba(df.to_numpy(), weights=weights),
        optimize_mode=optimize_mode,
    )

    logger.debug(
        "      * Done calculating distance matrix. Elapsed: %.1fs",
        time.time() - tstart,
    )
    return rtn



#!/usr/bin/env python3
"""Implement model domain functionality."""
import copy
import logging

import attr
import numpy as np
import pyproj

from .plots import DEF_FIGSHOW_CONFIG, get_domain_fig

logger = logging.getLogger(__name__)

# We'll approximate the Earth as a sphere
_EARTH_RADIUS = 6.3710088e6  # Mean earth radius in meters
_EQUATOR_PERIM = 2 * np.pi * _EARTH_RADIUS


@attr.s
class GridAxisConfig:
    """Configs for a grid axis."""

    # pylint: disable=locally-disabled, too-few-public-methods
    start = attr.ib()
    end = attr.ib()
    npts = attr.ib(validator=attr.validators.instance_of(int))
    npts_ezone = attr.ib(validator=attr.validators.instance_of(int), default=0)

    def __attrs_post_init__(self):
        if self.npts < 1:
            raise ValueError("npts must be larger than zero")
        if self.npts_ezone < 0:
            raise ValueError("npts_ezone must be positive")
        if self.start > self.end:
            self.start, self.end = self.end, self.start


class Grid2D:
    """A general grid in two dimensions, named x and y by convention."""

    def __init__(self, xaxis, yaxis, tstep=0.0):
        """Initialise params. Axis limits are open-ended intervals."""
        self._xaxis = self._validated_axis_info(xaxis)
        self._yaxis = self._validated_axis_info(yaxis)
        if tstep < 0:
            raise ValueError("tstep must be >= 0")
        self.tstep = tstep

        self.cart2grid = np.vectorize(self._cart2grid_non_vec)

    @staticmethod
    def _validated_axis_info(axis):
        if not isinstance(axis, GridAxisConfig):
            raise TypeError(
                "expected type 'GridAxisConfig', got '%s' instead"
                % (type(axis).__name__)
            )
        return axis

    @property
    def nx(self):
        """Return the number of grid points along the x-direction."""
        return self._xaxis.npts

    @nx.setter
    def nx(self, value):
        if value < 1:
            raise ValueError("nx must be larger than zero")
        self._xaxis.npts = value

    @property
    def nx_ezone(self):
        """Return number of extension zone grid pts along the x-direction."""
        return self._xaxis.npts_ezone

    @nx_ezone.setter
    def nx_ezone(self, value):
        if value < 0:
            raise ValueError("nx_ezone must be positive")
        self._xaxis.npts_ezone = value

    @property
    def ny(self):
        """Return the number of grid points along the y-direction."""
        return self._yaxis.npts

    @ny.setter
    def ny(self, value):
        if value < 1:
            raise ValueError("ny must be larger than zero")
        self._yaxis.npts = value

    @property
    def ny_ezone(self):
        """Return number of extension zone grid pts along the y-direction."""
        return self._yaxis.npts_ezone

    @ny_ezone.setter
    def ny_ezone(self, value):
        if value < 0:
            raise ValueError("ny_ezone must be positive")
        self._yaxis.npts_ezone = value

    @property
    def xmin(self):
        """Return the smallest value of x in the grid."""
        return self._xaxis.start

    @property
    def xmax(self):
        """Return the largest value of x in the grid."""
        return self._xaxis.end

    @property
    def xlims(self):
        """Return (xmin, xmax)."""
        return self.xmin, self.xmax

    @property
    def ymin(self):
        """Return the smallest value of y in the grid."""
        return self._yaxis.start

    @property
    def ymax(self):
        """Return the largest value of y in the grid."""
        return self._yaxis.end

    @property
    def ylims(self):
        """Return (ymin, ymax)."""
        return self.ymin, self.ymax

    @property
    def ezone_xmax(self):
        """Return the max value of x within the projected extension zone."""
        return self.xmax + self.nx_ezone * self.x_spacing

    @property
    def ezone_ymax(self):
        """Return the max value of y within the projected extension zone."""
        return self.ymax + self.ny_ezone * self.y_spacing

    @property
    def x_spacing(self):
        """Return the spacing, in meters, between nearest grid x-coords."""
        return (self.xmax - self.xmin) / self.nx

    @property
    def y_spacing(self):
        """Return the spacing, in meters, between nearest grid y-coords."""
        return (self.ymax - self.ymin) / self.ny

    @property
    def ncells(self):
        """Return the total number of grid cells."""
        return self.nx * self.ny

    def _cart2grid_non_vec(self, x, y):
        """Return grid (i, j) cell containing projected (x, y) coords."""
        i = int((x - self.xmin) / self.x_spacing)
        j = int((y - self.ymin) / self.y_spacing)
        if i < 0 or i > self.ncells - 1:
            i = None
        if j < 0 or j > self.ncells - 1:
            j = None
        return i, j

    def ij2xy_map(self):
        """Return grid2x, grid2y such that x=grid2x[i, j], y=grid2y[i, j]."""
        # We can then work with x=grid2x[i, j], y=grid2y[i, j]
        xvals = np.linspace(self.xmin, self.xmax, self.nx, endpoint=False)
        yvals = np.linspace(self.ymin, self.ymax, self.ny, endpoint=False)
        grid2x, grid2y = np.meshgrid(xvals, yvals)
        return grid2x, grid2y

    @property
    def corners(self):
        """Return tuple of projected (x, y) coords of the grid corners."""
        return (
            (self.xmin, self.ymin),
            (self.xmax, self.ymin),
            (self.xmax, self.ymax),
            (self.xmin, self.ymax),
        )

    @property
    def ezone_corners(self):
        """Return tuple of projected (x, y) coords of the ezone corners."""
        return (
            (self.xmax, self.ymin),
            (self.ezone_xmax, self.ymin),
            (self.ezone_xmax, self.ezone_ymax),
            (self.xmin, self.ezone_ymax),
            (self.xmin, self.ymax),
            (self.xmax, self.ymax),
        )


class DomainProjection:
    """Cartographic projection to be used in a domain."""

    def __init__(self, name="merc", lon0=0.0, lat0=0.0):
        """Initialise with passed attrs and add a transformer attr."""
        self.name = name
        self.lon0 = lon0
        self.lat0 = lat0
        self.transformer = pyproj.Proj(self.projparams)

    @property
    def full_name(self):
        """Return the full name of the used projection."""
        proj2projname = dict(
            merc="Mercator",
            stere="Polar Stereographic",
            lcc="Lambert Conformal Conic",
        )
        return proj2projname.get(self.name, self.name)

    @property
    def projparams(self):
        """Return the domain's projparams. See <https://proj.org>."""
        return dict(
            # Lambert conformal conic
            proj=self.name,
            R=_EARTH_RADIUS,
            lat_0=self.lat0,  # Lat of domain center. Is this maybe latc?
            lon_0=self.lon0,  # Lon of domain center. Is this maybe lonc?
            lat_1=self.lat0,  # 1st standard parallel
            lat_2=self.lat0,  # 2nd standard parallel
            # a, b: semimajor and semiminor axes (ellipsoidal planet)
            a=_EARTH_RADIUS,
            b=_EARTH_RADIUS,
        )

    def lonlat2xy(self, lon, lat):
        """Convert (lon, lat), in degrees, into projected (x, y) in meters."""
        # lat, lon in degrees
        # returns x, y in meters
        return self.transformer(longitude=lon, latitude=lat)

    def xy2lonlat(self, x, y):
        """Convert projected (x, y), in meters, into (lon, lat) in degrees."""
        lon, lat = self.transformer(x, y, inverse=True)
        return lon, lat


class DomainGrid(Grid2D):
    """A 2D grid with (lon, lat) <--> (x, y) projection awareness."""

    def __init__(self, *args, proj, **kwargs):
        """Initialise attrs from parent Grid2D class and add proj attr."""
        super().__init__(*args, **kwargs)
        self._proj = proj

    @property
    def proj(self):
        """Return the grid's associated DomainProjection object."""
        return self._proj

    @property
    def nlon(self):
        """Return the number of grid points along the longitude axis."""
        return self.nx

    @nlon.setter
    def nlon(self, value):
        self.nx = value

    @property
    def nlat(self):
        """Return the number of grid points along the latitude axis."""
        return self.ny

    @nlat.setter
    def nlat(self, value):
        self.ny = value

    def lonlat2grid(self, lon, lat):
        """Convert (lon, lat) into grid cell coord (i, j)."""
        x, y = self.proj.lonlat2xy(lon, lat)
        return self.cart2grid(x, y)

    def ij2lonlat_map(self):
        """Return g2lon, g2lat such that lon=g2lon[i, j], lat=g2lat[i, j]."""
        return self.proj.xy2lonlat(*self.ij2xy_map())

    @property
    def corners_lonlat(self):
        """Return tuple of (lon, lat) coords of the grid corners."""
        return tuple(self.proj.xy2lonlat(*xy) for xy in self.corners)

    @property
    def ezone_corners_lonlat(self):
        """Return tuple of (lon, lat) coords of the extension zone corners."""
        return tuple(self.proj.xy2lonlat(*xy) for xy in self.ezone_corners)


class Domain:
    """Model domain geometry and grid.

    See <https://hirlam.org/trac/wiki/HarmonieSystemDocumentation/ModelDomain>.

    """

    def __init__(
        self,
        name="",
        center_lonlat=(0.0, 0.0),
        proj_lon0_lat0=(0.0, 0.0),
        lmrt=False,
        ngrid_lonlat=(1, 1),
        grid_spacing=_EQUATOR_PERIM,
        ezone_ngrid=0,
        subdomain_split=(1, 1),
        coarse_grid_factor=0,
        tstep=0.0,
    ):
        """Initialise name, geometry, projection & grid/thinning grid attrs."""
        self.name = name

        # subdomain splits
        self._nsplit_lonlat = subdomain_split

        #########################
        # Initialise projection #
        #########################
        self.lmrt = lmrt

        def init_proj(ngrid_lonlat, proj_lon0_lat0, grid_spacing):
            """Help routine to initialise domain projection."""
            if self.lmrt and abs(proj_lon0_lat0[1]) > 0:
                logger.warning(
                    "lat0 should be 0 if lmrt=True. Resetting lat0 to 0."
                )
                proj_lon0_lat0 = (proj_lon0_lat0[0], 0.0)

            y_range = ngrid_lonlat[1] * grid_spacing
            return DomainProjection(
                name=self._auto_choose_projname(
                    lat0=proj_lon0_lat0[1], y_range=y_range
                ),
                lon0=proj_lon0_lat0[0],
                lat0=proj_lon0_lat0[1],
            )

        proj = init_proj(ngrid_lonlat, proj_lon0_lat0, grid_spacing)

        ###################
        # Initialise grid #
        ###################
        def init_grid(ngrid_lonlat, grid_spacing, ezone_ngrid):
            """Help routine to initialise domain grid."""
            # (a) Get projected coords of grid center
            center_xy = proj.lonlat2xy(
                lon=center_lonlat[0], lat=center_lonlat[1]
            )
            # (b) Grid x-axis
            x_range = ngrid_lonlat[0] * grid_spacing
            xmin = center_xy[0] - 0.5 * x_range
            xmax = center_xy[0] + 0.5 * x_range
            grid_xaxis = GridAxisConfig(
                start=xmin,
                end=xmax,
                npts=ngrid_lonlat[0],
                npts_ezone=ezone_ngrid,
            )
            # (c) Grid y-axis
            y_range = ngrid_lonlat[1] * grid_spacing
            ymin = center_xy[1] - 0.5 * y_range
            ymax = center_xy[1] + 0.5 * y_range
            grid_yaxis = GridAxisConfig(
                start=ymin,
                end=ymax,
                npts=ngrid_lonlat[1],
                npts_ezone=ezone_ngrid,
            )
            # (d) Set _grid attr
            return DomainGrid(
                xaxis=grid_xaxis, yaxis=grid_yaxis, proj=proj, tstep=tstep,
            )

        self._grid = init_grid(ngrid_lonlat, grid_spacing, ezone_ngrid)

        # Thinning grid
        if coarse_grid_factor < 1:
            self._thinning_grid = None
        else:
            # If you don't deepcopy here then you end up altering self.grid
            self._thinning_grid = copy.deepcopy(self.grid)
            self._thinning_grid.nx //= coarse_grid_factor
            self._thinning_grid.ny //= coarse_grid_factor
            self._thinning_grid.nx_ezone //= coarse_grid_factor
            self._thinning_grid.ny_ezone //= coarse_grid_factor

        ##############################################
        # Create vectorized versions of some methods #
        ##############################################
        self.contains_lonlat = np.vectorize(self._contains_lonlat_non_vec)

    @classmethod
    def construct_from_dict(cls, config):
        """Construct domain from info in the config dict.

        Args:
            config (netatmoqc.config.ConfigDict): Configs dictionary.

        Returns:
            netatmoqc.domains.Domain: New class instance constructed from
                info in the config dict.

        """
        return cls(
            name=config.name,
            ngrid_lonlat=(config.nlon, config.nlat),
            center_lonlat=(config.lonc, config.latc),
            proj_lon0_lat0=(config.lon0, config.lat0),
            grid_spacing=config.gsize,
            ezone_ngrid=config.ezone,
            lmrt=config.lmrt,
            subdomain_split=(config.split.lon, config.split.lat),
            coarse_grid_factor=config.thinning_grid_coarse_factor,
            tstep=config.tstep,
        )

    def _auto_choose_projname(self, lat0, y_range):
        """Define domain projection."""
        # Do this in a way close to what is explained at
        # <https://hirlam.org/trac/wiki/HarmonieSystemDocumentation/ModelDomain>

        latrange = 180 * y_range / _EQUATOR_PERIM
        if self.lmrt or latrange > 35 or np.isclose(latrange, 0):
            # <https://proj.org/operations/projections/merc.html>
            # <https://desktop.arcgis.com/en/arcmap/10.3/guide-books/
            #  map-projections/mercator.htm>
            rtn = "merc"
        elif np.isclose(lat0, 90.0):
            # <https://proj.org/operations/projections/stere.html>
            # <https://desktop.arcgis.com/en/arcmap/10.3/guide-books/
            #  map-projections/polar-stereographic.htm>
            rtn = "stere"
        else:
            # <https://proj.org/operations/projections/lcc.html>
            # <https://desktop.arcgis.com/en/arcmap/10.3/guide-books/
            #  map-projections/lambert-conformal-conic.htm>
            rtn = "lcc"

        return rtn

    @property
    def grid(self):
        """Return the grid used in the domain discretisation."""
        return self._grid

    @property
    def thinning_grid(self):
        """Return the grid used for thinning of observations in the domain."""
        return self._thinning_grid

    @property
    def proj(self):
        """Return the domain's associated DomainProjection object."""
        # Make sure the domain and the domain's grid use the same projection
        return self.grid.proj

    @property
    def center_lonlat(self):
        """Return the (lon, lat) coords of the domain center."""
        center_x = 0.5 * (self.grid.xmin + self.grid.xmax)
        center_y = 0.5 * (self.grid.ymin + self.grid.ymax)
        return self.proj.xy2lonlat(center_x, center_y)

    @property
    def n_subdomains(self):
        """Return the #subdomains given by self.split with default args."""
        return np.prod(self._nsplit_lonlat)

    def split(self, nsplit_lon=None, nsplit_lat=None):
        """Split a domain "nsplit_lon(lat)" times along the lon(lat) axis.

        Args:
            nsplit_lon (int): Number of divisions along the longitude axis.
                (Default value = None).
            nsplit_lat (int): Number of divisions along the latitude axis.
                (Default value = None).

        Returns:
            list: List of subdomains (Domain objects) obtained as a result of
                the split.

        Raises:
            TypeError: If nsplit_lon, nsplit_lon are not integers.
            ValueError: If nsplit_lon and nsplit_lat do not divide
                nlon and nlat, respectively.

        """
        if nsplit_lon is None:
            nsplit_lon = self._nsplit_lonlat[0]
        if nsplit_lat is None:
            nsplit_lat = self._nsplit_lonlat[1]

        if not isinstance(nsplit_lon, int):
            raise TypeError("'nsplit_lon' must be an integer")
        if not isinstance(nsplit_lat, int):
            raise TypeError("'nsplit_lat' must be an integer")
        if nsplit_lon <= 0:
            raise ValueError("'nsplit_lon' must be larger than zero")
        if nsplit_lat <= 0:
            raise ValueError("'nsplit_lat' must be larger than zero")
        if self.grid.nlon % nsplit_lon != 0:
            raise ValueError("'nsplit_lon' must divide nlon")
        if self.grid.nlat % nsplit_lat != 0:
            raise ValueError("'nsplit_lat' must divide nlat")

        new_nlon = self.grid.nlon // nsplit_lon
        new_nlat = self.grid.nlat // nsplit_lat
        nsplit = nsplit_lon * nsplit_lat

        splits = []
        for ilat in range(nsplit_lat):
            new_yc = (
                self.grid.ymin + (ilat + 0.5) * new_nlat * self.grid.y_spacing
            )
            for ilon in range(nsplit_lon):
                new_xc = (
                    self.grid.xmin
                    + (ilon + 0.5) * new_nlon * self.grid.x_spacing
                )
                new_lonc, new_latc = self.proj.xy2lonlat(new_xc, new_yc)
                split = self.__class__(
                    name="{} Split {}/{}".format(
                        self.name, ilat * nsplit_lat + ilon + 1, nsplit
                    ),
                    ngrid_lonlat=(new_nlon, new_nlat),
                    center_lonlat=(new_lonc, new_latc),
                    proj_lon0_lat0=(self.proj.lon0, self.proj.lat0),
                    # We're using same spacing along x and y
                    grid_spacing=self.grid.x_spacing,
                    # We don't want the splits to have extension zones
                    ezone_ngrid=0,
                    lmrt=self.lmrt,
                    tstep=self.grid.tstep,
                )
                splits.append(split)
        return splits

    def trim_obs(self, df):
        """Remove from dataframe "df" those obs that lie outside the domain.

        Args:
            df (pandas.Dataframe): Dataframe to be trimmed.

        Returns:
            pandas.Dataframe: Trimmed dataframe.

        """
        if len(df.index) == 0:
            return df
        trimmed_df = df[self.contains_lonlat(df.lon, df.lat)].copy()
        trimmed_df = trimmed_df.reset_index(drop=True)
        return trimmed_df

    @property
    def minlat(self):
        """Return the min latitude of the domain's corners."""
        return min(lonlat[1] for lonlat in self.grid.corners_lonlat)

    @property
    def minlon(self):
        """Return the min longitude of the domain's corners."""
        return min(lonlat[0] for lonlat in self.grid.corners_lonlat)

    @property
    def maxlat(self):
        """Return the max latitude of the domain's corners."""
        return max(lonlat[1] for lonlat in self.grid.corners_lonlat)

    @property
    def maxlon(self):
        """Return the max longitude of the domain's corners."""
        return max(lonlat[0] for lonlat in self.grid.corners_lonlat)

    @property
    def ezone_maxlat(self):
        """Return the max latitude of the domain's ezone corners."""
        return max(lonlat[1] for lonlat in self.grid.ezone_corners_lonlat)

    @property
    def ezone_maxlon(self):
        """Return the max longitude of the domain's ezone corners."""
        return max(lonlat[0] for lonlat in self.grid.ezone_corners_lonlat)

    @property
    def ezone_minlat(self):
        """Return the min latitude of the domain's ezone corners."""
        return min(lonlat[1] for lonlat in self.grid.ezone_corners_lonlat)

    @property
    def ezone_minlon(self):
        """Return the min longitude of the domain's ezone corners."""
        return min(lonlat[0] for lonlat in self.grid.ezone_corners_lonlat)

    def _contains_lonlat_non_vec(self, lon, lat):
        """Return True if (lon, lat) is within domain, and False otherwise."""
        x, y = self.proj.lonlat2xy(lon, lat)
        return (self.grid.xmin <= x < self.grid.xmax) and (
            self.grid.ymin <= y < self.grid.ymax
        )

    def get_fig(self, **kwargs):
        """Get domain figure.

        Args:
            **kwargs: Passed to get_domain_fig.

        Returns:
            go.Figure: Domain figure.

        """
        return get_domain_fig(self, **kwargs)

    def show(self, config=None, **kwargs):
        """Show domain figure.

        Args:
            config (dict): Plotly's fig.show config, not our parsed config
                file.
            **kwargs: Passed to self.get_fig.

        """
        default_config = DEF_FIGSHOW_CONFIG.copy()
        if config is not None:
            default_config.update(config)
        config = default_config

        fig = self.get_fig(**kwargs)
        fig.show(config=config)



#!/usr/bin/env python3
"""Web Server Gateway Interface."""
# pylint: disable=unused-import
# See <https://community.plotly.com/t/multiple-dashboards/4656/3>
import clustering.app
import scattergeo_timeseries.app
from server import server as application



#!/usr/bin/env python3
"""Parse the config file."""
import copy
import functools
import io
import logging
import os
import sys
import tempfile
from contextlib import contextmanager
from pathlib import Path

import numpy as np
import toml
from dotmap import DotMap

from .dtgs import DtgContainer

logger = logging.getLogger(__name__)

###############################
# General types and functions #
###############################


class UndefinedValueType:
    """UndefinedValueType class.

    This class is defined such that any instance foo satisfies:
        (a) bool(foo) is False
        (b) If return_self_on_attr_error=True, then, for any attribute bar,
            foo.bar is foo. This implies that, for any N>1,
            foo.bar1.bar2.(...).barN is foo

    We will use instances of this class to mark things such as missing
    config file values, unspecified defaults in functions, etc. We don't
    want to use None in these cases, as None may be a valid arg value.

    """

    def __init__(
        self,
        name="<UndefinedValueType object>",
        return_self_on_attr_error=False,
    ):
        """Set instance name and return_self_on_attr_error attr."""
        self._name = name
        self._return_self_on_attr_error = return_self_on_attr_error

    def __setitem__(self, item, val):
        pass

    def __getitem__(self, item):
        return self

    # Define __setstate__ and __getstate__ to handle serialisation (pickling)
    # and avoid recursion errors (UndefinedConfigValue is a recursive object)
    def __setstate__(self, state):
        self.__dict__ = state

    def __getstate__(self):
        return self.__dict__

    def __getattr__(self, item):
        if self._return_self_on_attr_error:
            return self
        raise AttributeError(
            "'{}' object has no attribute '{}'".format(
                self.__class__.__name__, item
            )
        )

    def __copy__(self):
        return self

    def __deepcopy__(self, memo):
        return self

    def __bool__(self):
        return False

    def __repr__(self):
        return self._name

    __str__ = __repr__


UndefinedConfigValue = UndefinedValueType(
    name="UndefinedConfigValue", return_self_on_attr_error=True
)

NoDefaultProvided = UndefinedValueType(name="NoDefaultProvided")


class UndefinedConfigValueError(Exception):
    """Exception to be raised if UndefinedConfigValue is encountered."""


class ConfigDict(DotMap):
    """Config dictionary.

    Acts like a normal dictionary, except for:
        (a) Accepts getting/setting values using dot notation (inherited from
            the dotmap package). This means that instance.key is equivalent
            to instance[key].
        (b) Accepts getting/setting values using "hierarchy lists", in the
            sense that the retrievals
                        instance[['at1', 'at2', (...), 'atn']],
                        instance[['at1.at2.(...).atn']],
                        instance['at1.at2.(...).atn'],
                        instance.at1.at2.(...).atn
            are equivalent.
        (c) If self._dynamic is True, then, instead of raising KeyError:
            * Retrievals on missing keys return UndefinedConfigValue
            * Sublevels are automatically created if needed upon item setting

    """

    def __init__(self, *args, **kwargs):
        """Make KeyError possible if "_dynamic=True" not explicitly passed."""
        if "_dynamic" not in kwargs:
            kwargs["_dynamic"] = False
        super().__init__(*args, **kwargs)

    def get(self, key, default=UndefinedConfigValue):
        """Implement ".get". Default UndefinedConfigValue instead of None."""
        if not isinstance(key, str):
            # Implement retrievals using "hierarchy lists"
            try:
                key = ".".join(iter(key))
            except (TypeError) as err:
                msg = (
                    "The attribute(s) to be retrieved must be specified "
                    + "either as a string or as an iterable type. "
                    + 'Got "{}" instead (type={})'
                ).format(key, type(key))
                raise TypeError(msg) from err

        def _getattr(obj, key, default=default):
            if isinstance(obj, type(self)):
                try:
                    rtn = super().__getitem__(key)
                    if rtn == type(self)():
                        raise KeyError(key)
                except KeyError:
                    if default is NoDefaultProvided:
                        raise
                    rtn = default
            else:
                try:
                    rtn = getattr(obj, key)
                except AttributeError:
                    rtn = default
            return rtn

        return functools.reduce(_getattr, [self] + key.split("."))

    def __getitem__(self, attr):
        if self._dynamic:
            rtn = self.get(attr, default=UndefinedConfigValue)
        else:
            rtn = self.get(attr, default=NoDefaultProvided)
        return rtn

    def __setitem__(self, attr, val):
        if "." in attr:
            pre, _, post = attr.rpartition(".")
            if self.get(pre) is UndefinedConfigValue:
                if self._dynamic:
                    new_pre = type(self)()
                    new_pre[post] = UndefinedConfigValue
                    new_pre._dynamic = True
                    setattr(self, pre, new_pre)
                else:
                    raise KeyError(attr)
            setattr(self.get(pre) if pre else self, post, val)
        else:
            super().__setitem__(attr, val)

    def set_dynamic_flags(self, boolean):
        """Recursively set "_dynamic" to True/False."""
        boolean = bool(boolean)
        self._dynamic = boolean
        for key, val in self.items():
            if isinstance(val, type(self)):
                val.set_dynamic_flags(boolean)
                self[key] = val
                self[key]._dynamic = boolean
            else:
                pass

    def __copy__(self):
        cls = self.__class__
        result = cls.__new__(cls, _dynamic=self._dynamic)
        result.__dict__.update(self.__dict__)
        return result

    def __deepcopy__(self, memo):
        cls = self.__class__
        result = cls.__new__(cls, _dynamic=self._dynamic)
        memo[id(self)] = result
        for key, val in self.__dict__.items():
            setattr(result, key, copy.deepcopy(val, memo))
        return result

    def toml_dumps(self):
        """Dump config in TOML format."""
        return toml.dumps(self.toDict(), encoder=toml.TomlPathlibEncoder(dict))

    def __str__(self):
        orig_stdout = sys.stdout
        tmp_stdout = io.StringIO()
        sys.stdout = tmp_stdout
        self.pprint()
        output = tmp_stdout.getvalue()
        sys.stdout = orig_stdout
        return output


######################################################
# Machinery to allow registering config opt metadata #
######################################################


class _MetadataDict(ConfigDict):
    pass


class _ConfigMetadataRegistry(ConfigDict):
    """Class to help register metadata about known config options."""

    def register(
        self,
        key,
        default=NoDefaultProvided,
        minval=NoDefaultProvided,
        maxval=NoDefaultProvided,
        choices=NoDefaultProvided,
        astype=NoDefaultProvided,
        case_sensitive=NoDefaultProvided,
    ):
        """Register a metadata registry entry.

        Args:
            key (str): The key (name) of the param metadata to be registed.
            default (obj):  The default value for the registered metadata.
                (Default value = NoDefaultProvided)
            minval (obj): An optional min value.
                (Default value = NoDefaultProvided).
            maxval (obj): An optional max value.
                (Default value = NoDefaultProvided).
            choices (list): Optional list of allowed choices.
                (Default value = NoDefaultProvided).
            astype (type): Optional type info for the metadata. Used, e.g.,
                for typecasting. (Default value = NoDefaultProvided).
            case_sensitive (bool): Whether future lookups should consider
                character case in the metadata key.
                (Default value = NoDefaultProvided).

        Raises:
            TypeError: If key is not a string.

        """
        if not isinstance(key, str):
            raise TypeError("key must be a string")
        try:
            # key -> section.key.
            # "section" will be defined using a context manager
            key = "{}.{}".format(section, key)
        except NameError:
            pass

        if astype is NoDefaultProvided and default is not NoDefaultProvided:
            astype = type(default)
        self._dynamic = True
        metadata = _MetadataDict(
            default=default,
            minval=minval,
            maxval=maxval,
            choices=choices,
            astype=astype,
            case_sensitive=case_sensitive,
        )
        metadata._dynamic = False
        self[key] = metadata
        self._dynamic = False

    def copy_template(self, key):
        """Copy contents from self["_template"] to section.key.

        "section" needs to be defined in the upper-level scope,
        using a context manager.

        Args:
            key (str): String such that the destination where the template
                will be copied into is section.key.
        """
        dest_key = "{}.{}".format(section, key)
        if "." in section:
            template_key, _, _dropped = section.rpartition(".")
            template_key = "_template." + template_key + "." + key
        else:
            template_key = "_template." + section + "." + key
        self._dynamic = True
        self[dest_key] = copy.deepcopy(self[template_key])
        self._dynamic = False


config_metadata = _ConfigMetadataRegistry(_dynamic=False)


@contextmanager
def config_section(sec):
    """Yield a config section name for use as context manager."""
    yield sec


###########################################################################
# Registering config options (known opts, defaults, restrictions, etc).   #
# This is where you should change, e.g., in order to add new config opts. #
###########################################################################

# Create a template for filling metadata for the various clustering_methods
with config_section("_template.clustering_method") as section:
    config_metadata.register("eps", default=10.0, minval=0.0)
    config_metadata.register("min_samples", default=5, minval=1)
    config_metadata.register("min_cluster_size", default=5, minval=1)
    # obs_weights not explicitly set will be internally set to 1
    config_metadata.register("obs_weights.geo_dist", default=1.0, minval=0.0)
    config_metadata.register("obs_weights.alt", default=1.0, minval=0.0)
    config_metadata.register(
        "obs_weights.temperature", default=5.0, minval=0.0
    )
    config_metadata.register("obs_weights.pressure", default=0.0, minval=0.0)
    config_metadata.register("obs_weights.mslp", default=1.0, minval=0.0)
    config_metadata.register("obs_weights.humidity", default=1.0, minval=0.0)
    config_metadata.register("obs_weights.sum_rain_1", default=1.0, minval=0.0)
    config_metadata.register(
        "outlier_removal.method",
        default="iterative",
        choices=[None, "lof", "iterative", "reclustering"],
    )
    config_metadata.register(
        "outlier_removal.iterative.max_n_stdev", default=3.0, minval=0.0
    )
    config_metadata.register(
        "outlier_removal.iterative.max_n_iter", default=100, minval=0
    )


# General section
def parsed_path(path):
    """Return path with envvars expanded and as a pathlib.Path obj."""
    return Path(os.path.expandvars(str(path)))


with config_section("general") as section:
    # in/our dirs
    config_metadata.register(
        "data_rootdir", default=".", astype=parsed_path,
    )
    config_metadata.register(
        "outdir",
        default=Path(tempfile.gettempdir()).resolve() / "netatmoqc_output",
        astype=parsed_path,
    )
    # Chosen clustering method
    config_metadata.register(
        "clustering_method",
        default="hdbscan",
        choices=["hdbscan", "dbscan", "optics", "rsl"],
    )
    # DTGs
    config_metadata.register("dtgs.list")
    config_metadata.register("dtgs.start")
    config_metadata.register("dtgs.end")
    config_metadata.register("dtgs.cycle_length", default="3H")
    # Metrics optimisaion scheme
    config_metadata.register(
        "custom_metrics_optimize_mode",
        default="memory",
        choices=["speed", "memory", "speed_mem_compromise"],
    )
    # Data cols to export when saving obsoul output
    config_metadata.register(
        "obsoul_export_params",
        default=["pressure", "temperature", "humidity", "sum_rain_1"],
    )

# "commans section"
with config_section("commands.select") as section:
    config_metadata.register("station_rejection_tol", default=0.15, minval=0.0)

# clustering_method.hdbscan
with config_section("clustering_method.hdbscan") as section:
    config_metadata.copy_template("min_samples")
    config_metadata.copy_template("min_cluster_size")
    config_metadata.copy_template("obs_weights")
    config_metadata.register(
        "outlier_removal.method",
        default="glosh",
        choices=[None, "glosh", "lof", "iterative", "reclustering"],
    )
    config_metadata.copy_template("outlier_removal.iterative")

# clustering_method.dbscan
with config_section("clustering_method.dbscan") as section:
    config_metadata.copy_template("eps")
    config_metadata.copy_template("min_samples")
    config_metadata.copy_template("obs_weights")
    config_metadata.copy_template("outlier_removal")

# clustering_method.optics
with config_section("clustering_method.optics") as section:
    config_metadata.copy_template("min_samples")
    config_metadata.copy_template("min_cluster_size")
    config_metadata.copy_template("obs_weights")
    config_metadata.copy_template("outlier_removal")

# clustering_method.rsl
with config_section("clustering_method.rsl") as section:
    # eps -> cut
    config_metadata.copy_template("eps")
    # min_samples -> k
    config_metadata.copy_template("min_samples")
    # min_cluster_size -> gamma
    config_metadata.copy_template("min_cluster_size")
    config_metadata.copy_template("obs_weights")
    config_metadata.copy_template("outlier_removal")

# domain
with config_section("domain") as section:
    config_metadata.register("name", default="")
    # split the main grid in small grids containing
    # split.lon longitudes and split.lat latitudes
    # for a first clustering. Quicker and saves memory.
    config_metadata.register("split.lon", default=3, minval=1, astype=int)
    config_metadata.register("split.lat", default=4, minval=1, astype=int)
    # These defaults are for the METCOOP25C domain. See
    # <https://hirlam.org/trac/wiki/HarmonieSystemDocumentation/ModelDomain>
    # <https://hirlam.org/trac/browser/Harmonie/scr/Harmonie_domains.pm>
    config_metadata.register("tstep", default=75, minval=0, astype=int)
    config_metadata.register("nlon", default=900, minval=1, astype=int)
    config_metadata.register("nlat", default=960, minval=1, astype=int)
    config_metadata.register(
        "lonc", default=16.763011639, minval=-180, maxval=180
    )
    config_metadata.register(
        "latc", default=63.489212956, minval=-90, maxval=90
    )
    config_metadata.register("lon0", default=15.0, minval=-180, maxval=180)
    config_metadata.register("lat0", default=63.0, minval=-90, maxval=90)
    config_metadata.register("gsize", default=2500.0, minval=0)
    config_metadata.register("ezone", default=11, minval=0, astype=int)
    config_metadata.register("lmrt", default=False, choices=[True, False])
    # At the end of its run, the "select" command uses a coarser version of the
    # domain's grid to thin the returned data (at most one obs per coarse grid
    # point is returned). The config param "thinning_grid_coarse_factor"
    # controls how coarser this coarse grid will be.
    config_metadata.register(
        "thinning_grid_coarse_factor", default=4, minval=0, astype=int
    )


#####################################
# Machinery to parse config options #
#####################################


def _parse_dtg_entires(dtgs_config):
    try:
        cycle_length = dtgs_config["cycle_length"]
    except KeyError:
        cycle_length = "3H"
        logger.warning("Missing cycle_length in config. Setting it to 3H.")

    dtg_range_opt_labels = {"start", "end"}
    if "list" in dtgs_config.keys():
        if not dtg_range_opt_labels.isdisjoint(dtgs_config.keys()):
            logger.warning(
                (
                    'Found DTG options "list" and also at least '
                    'one of [%s]. Parsing only values passed in "list".'
                ),
                ", ".join(dtg_range_opt_labels),
            )
        try:
            dtgs = DtgContainer(dtgs_config["list"], cycle_length=cycle_length)
        except ValueError as err:
            raise ValueError(
                "Bad 'dtg.list' or 'dtg.cycle_length' config."
            ) from err
    else:
        try:
            _ = dtgs_config["start"]
            _ = dtgs_config["end"]
        except (KeyError) as err:
            msg = " Either 'dtgs.list' or both 'dtgs.start' and "
            msg += "'dtgs.end' must specified in config file."
            raise UndefinedConfigValueError(msg) from err
        try:
            dtgs = DtgContainer(
                start=dtgs_config["start"],
                end=dtgs_config["end"],
                cycle_length=cycle_length,
            )
        except (ValueError, TypeError) as err:
            raise ValueError(
                "Bad 'dtg.start/end/cycle_length' config."
            ) from err

    return dtgs


def _raw2parsed(raw, recog_configs=config_metadata, parent_keys=()):
    """Return parsed confif from raw input, validating with metadata.

    Descend recursively into input dict raw and validate raw configs against
    the registered metadata. Configs not that don't have corresponding metadata
    info will be passed unchanged.

    Args:
        raw (dict): Raw configs, as read from config file.
        recog_configs (_ConfigMetadataRegistry): Configs for which we have
            registered metadata. Default value = config_metadata.
        parent_keys (tuple): Tuple of all config dict keys, in order, traversed
            to get to the current raw dict. Default value = (). Used for dicts
            nested within the top-level raw.

    Returns:
        ParsedConfig: Parsed configs from raw.

    Raises:
        ValueError: If any config in raw is not a valid choice, or has an
            invalid value.

    """
    # Do not use mutable data structures for argument defaults (parent_keys)
    parent_keys = list(parent_keys)
    parsed = ConfigDict(_dynamic=True)
    for key, user_val in raw.items():
        # We'll force all keys in the parsed dict to be strings
        key = str(key)
        all_keys = parent_keys + [key]
        full_key = ".".join(all_keys)
        metadata = recog_configs.get(full_key)

        if metadata is UndefinedConfigValue:
            # Leave as is
            logger.warning(
                "Config opt '%s' not recognised. Passed as is.", full_key
            )
            parsed[key] = user_val
            continue

        # Cast to metadata type, if any
        if metadata["astype"] and user_val is not None:
            try:
                user_val = metadata["astype"](user_val)
            except TypeError:
                logger.warning(
                    (
                        "Could not cast config opt '%s=%s' (type '%s') "
                        "into expected type '%s'. Passing without casting."
                    ),
                    full_key,
                    user_val,
                    type(user_val).__name__,
                    metadata["astype"].__name__,
                )

        if isinstance(user_val, dict):
            # Top level of some config section
            # Descend into further eventual dict levels recusively
            parsed[key] = _raw2parsed(user_val, parent_keys=all_keys)
        else:
            # Check if allowed choices (if any) are respected
            if metadata["choices"]:
                if not metadata["case_sensitive"]:
                    # Make it all lowercase
                    metadata["choices"] = [
                        c.lower() if isinstance(c, str) else c
                        for c in metadata["choices"]
                    ]
                    if isinstance(user_val, str):
                        user_val = user_val.lower()
                if user_val not in metadata["choices"]:
                    raise ValueError(
                        (
                            "In config opt '{}': '{}' is not a valid choice. "
                            + "Please choose from: {}"
                        ).format(
                            full_key,
                            user_val,
                            ", ".join(map(str, metadata["choices"])),
                        )
                    )
            # Check value bounds, if specified
            for bound, fun in zip(["minval", "maxval"], [np.less, np.greater]):
                if metadata[bound] is not NoDefaultProvided and fun(
                    user_val, metadata[bound]
                ):
                    raise ValueError(
                        (
                            "In config opt '{}': Value '{}' not allowed. "
                            + "The allowed {} is {}"
                        ).format(full_key, user_val, bound, metadata[bound])
                    )

            # Save parsed value
            parsed[key] = user_val

    parsed._dynamic = False
    return parsed


def _fill_defaults(raw, recog_configs=config_metadata, parent_keys=()):
    """Fill defaults into raw config according to registered metadata.

    Descend recursively into input dict raw and fill in default values
    according to metadata registered in this module.

    Args:
        raw (dict): Raw configs, as read from config file.
        recog_configs (_ConfigMetadataRegistry): Configs for which we have
            registered metadata. Default value = config_metadata.
        parent_keys (tuple): Tuple of all config dict keys, in order, traversed
            to get to the current raw dict. Default value = (). Used for dicts
            nested within the top-level raw.

    Returns:
        dict: Copy of raw with defaults filled in.

    Raises:
        TypeError: If any metadata dict found while parsing raw is not an
            instance of either _MetadataDict or ConfigDict classes.

    """
    # Do not use mutable data structures for argument defaults (parent_keys)
    parent_keys = list(parent_keys)
    parsed = ConfigDict(raw, _dynamic=False)
    for key, metadata in recog_configs.items():
        if key == "_template":
            # We don't want to include the template in the final config
            continue
        all_keys = parent_keys + [key]
        full_key = ".".join(all_keys)
        if isinstance(metadata, _MetadataDict):
            # We should not descend recursively into this
            try:
                user_value = parsed[key]
            except KeyError:
                if metadata.default is NoDefaultProvided:
                    logger.debug(
                        "Missing config opt '%s' with no default", full_key
                    )
                else:
                    logger.debug(
                        "Filling defaults for config opt '%s'", full_key
                    )
                    parsed[key] = metadata.default
                    if metadata.astype:
                        parsed[key] = metadata.astype(parsed[key])
        elif isinstance(metadata, ConfigDict):
            if len(metadata) != 0:
                try:
                    user_value = parsed[key]
                except KeyError:
                    user_value = ConfigDict()
                parsed[key] = _fill_defaults(user_value, metadata, all_keys)
        else:
            raise TypeError(
                "Expected only _MetadataDict or ConfigDict, got %s"
                % (type(metadata))
            )
    return parsed


##########################################################################
# Main objects that the rest of the package will primarily interact with #
##########################################################################


class ParsedConfig:
    """Parsed version of config file or raw dict.

    Each section in config will be accessible as self.section.

    """

    def __init__(
        self,
        arg0=Path().resolve() / "config.toml",
        global_default=NoDefaultProvided,
    ):
        """Initialise "file", "global_default", "raw", and "parsed" attrs."""
        # raw data input
        if isinstance(arg0, dict):
            self._file = None
            self._raw = ConfigDict(arg0, _dynamic=False)
        else:
            self._file = Path(arg0).resolve()
            # Keep the config's raw data, mostly for debugging purposes
            self._raw = ConfigDict(toml.load(self._file), _dynamic=False)

        self.global_default = global_default

        self._parsed = _fill_defaults(_raw2parsed(self._raw))

        # DTGs are treated a bit specially
        self._parsed.general.dtgs = _parse_dtg_entires(
            self._parsed.general.dtgs
        )

        self._parsed.set_dynamic_flags(False)

    # Define __setstate__ and __getstate__ to handle serialisation (pickling)
    # and avoid recursion errors
    def __setstate__(self, state):
        self.__dict__ = state

    def __getstate__(self):
        return self.__dict__

    # Handling the getter method
    # __getattr__ is only called if the attribute does NOT exist
    def __getattr__(self, item):
        return self.get(item, default=NoDefaultProvided)

    def get_clustering_opt(self, item, default=UndefinedConfigValue):
        """Get item from self.clustering_method for configd cluster method."""
        keys_for_get = ["clustering_method", self.general.clustering_method]
        keys_for_get.append(item)
        return self.get(keys_for_get, default)

    def get(self, item, default=UndefinedConfigValue):
        """Implement get method with default UndefinedConfigValue."""
        rtn = self._parsed.get(item, default=UndefinedConfigValue)
        if rtn is UndefinedConfigValue:
            if default is NoDefaultProvided:
                if self.global_default is NoDefaultProvided:
                    raise UndefinedConfigValueError(
                        'Config has no value for "{}", '.format(item)
                        + "and no default has been passed."
                    )

                logger.debug(
                    'No "%s" in config. Using global_default (= %s)',
                    item,
                    self.global_default,
                )
                rtn = self.global_default
            else:
                logger.debug(
                    'No "%s" in config. Using passed default (= %s)',
                    item,
                    default,
                )
                rtn = default

        return rtn

    def toml_dumps(self):
        """Dump config in TOML format."""
        parsed_copy = copy.deepcopy(self._parsed)
        parsed_copy.general.dtgs = self._raw.general.dtgs
        return parsed_copy.toml_dumps()

    def __str__(self):
        parsed_copy = copy.deepcopy(self._parsed)
        parsed_copy.general.dtgs = self._raw.general.dtgs
        return str(parsed_copy)


def read_config(config_path):
    """Read config file at location "config_path".

    Args:
        config_path (pathlib.Path): The path to the config file.

    Returns:
        ParsedConfig: Parsed configs read from config_path.

    """
    logging.info("Reading config file %s", Path(config_path).resolve())
    config = ParsedConfig(config_path, global_default=UndefinedConfigValue)
    return config



#!/usr/bin/env python3
"""Logging-related classes, functions and definitions."""
import logging
from collections import namedtuple

# Define aliases to ANSI escape sequences to set text color in log
LogColor = namedtuple(
    typename="LogColor",
    field_names=("reset", "yellow", "red", "green", "cyan"),
)

logcolor = LogColor(
    reset="\u001b[0m",
    yellow="\u001b[33m",
    red="\u001b[31m",
    green="\u001b[32m",
    cyan="\u001b[36m",
)


class CustomFormatter(logging.Formatter):
    """Logging Formatter to add colors and count warning / errors."""

    # Adapted from: <https://stackoverflow.com/q/14844970/3644857>

    FORMATS = {
        "DEFAULT": "%(asctime)s: %(message)s",
        logging.CRITICAL: (
            logcolor.red
            + "%(asctime)s %(levelname)s(%(module)s: %(lineno)d): %(message)s"
            + logcolor.reset
        ),
        logging.ERROR: logcolor.red
        + "%(asctime)s %(levelname)s: "
        + logcolor.reset
        + "%(message)s",
        logging.WARNING: logcolor.yellow
        + "%(asctime)s %(levelname)s: "
        + logcolor.reset
        + "%(message)s",
        logging.DEBUG: (
            logcolor.green
            + "%(asctime)s %(levelname)s(%(module)s: %(lineno)d): "
            + logcolor.reset
            + "%(message)s"
        ),
    }

    def format(self, record):  # noqa: A003 (class attribute shadowing builtin)
        """Return a formatter.format for "record"."""
        log_fmt = self.FORMATS.get(record.levelno, self.FORMATS["DEFAULT"])
        formatter = logging.Formatter(log_fmt, datefmt="%Y-%m-%d %H:%M:%S")
        return formatter.format(record)


def get_logger(name, loglevel):
    """Get logger with name "name" and loglevel "loglevel"."""
    logger = logging.getLogger(name)
    logger_handler = logging.StreamHandler()
    logger_handler.setFormatter(CustomFormatter())
    logging.basicConfig(
        level=logging.getLevelName(loglevel.upper()),
        handlers=[logger_handler],
    )
    return logger



#!/usr/bin/env python3
"""Code for the scattergeo_timeseries app."""
import logging
import os
from datetime import datetime as dt

import dash
import dash_core_components as dcc
import dash_html_components as html
import numpy as np
import pandas as pd
import plotly.graph_objects as go
from dash.dependencies import Input, Output, State
from server import server

from netatmoqc.config_parser import read_config
from netatmoqc.domains import Domain
from netatmoqc.load_data import (
    read_netatmo_data_for_dtg,
    remove_irregular_stations,
)
from netatmoqc.logs import CustomFormatter
from netatmoqc.plots import generate_single_frame, init_fig_dict

logger = logging.getLogger(__name__)
logger_handler = logging.StreamHandler()
logger_handler.setFormatter(CustomFormatter())
logging.basicConfig(level=logging.INFO, handlers=[logger_handler])

config = read_config(os.getenv("NETATMOQC_CONFIG_PATH", "config.toml"))
domain = Domain.construct_from_dict(config.domain)

app = dash.Dash(
    name="scattergeo_timeseries",
    server=server,
    url_base_pathname="/scattergeo_timeseries/",
    meta_tags=[
        {"name": "viewport", "content": "width=device-width, initial-scale=1"}
    ],
)


def description_card():
    """Return a Div containing dashboard title & descriptions."""
    return html.Div(
        id="description-card",
        children=[
            html.H5("NetAtmo Data Explorer"),
            html.H3("NetAtmo Data Explorer Dashboard"),
            html.Div(id="intro", children="An aid to explore NetAtmo data",),
        ],
        style={"text-align": "center"},
    )


variable_list = [
    "temperature",
    "pressure",
    "mslp",
    "alt",
    "humidity",
    "sum_rain_1",
]


def generate_control_card():
    """Return: A Div containing controls for graphs."""
    return html.Div(
        id="control-card",
        children=[
            html.P("Select variable"),
            dcc.Dropdown(
                id="variable-select",
                options=[{"label": i, "value": i} for i in variable_list],
                value=variable_list[0],
            ),
            html.Br(),
            html.P("Select time range"),
            dcc.DatePickerRange(
                id="date-picker-select",
                minimum_nights=0,
                start_date=dt(2018, 4, 1),
                end_date=dt(2018, 4, 1),
                min_date_allowed=dt(2018, 4, 1),
                max_date_allowed=dt.today(),
                initial_visible_month=dt(2018, 4, 1),
            ),
            html.Br(),
            html.Br(),
            html.Div(
                id="plot-btn-outer",
                children=html.Button(
                    id="plot-btn", children="Make Plot", n_clicks=0
                ),
            ),
        ],
    )


app.layout = html.Div(
    id="app-container",
    children=[
        # Left column
        html.Div(
            id="left-column",
            className="three columns",
            children=[description_card(), generate_control_card()]
            + [
                html.Div(
                    ["initial child"],
                    id="output-clientside",
                    style={"display": "none"},
                )
            ],
        ),
        # Right column
        html.Div(
            id="right-column",
            className="nine columns",
            children=[
                html.Div(
                    id="scattergeo_plot_div",
                    children=[
                        html.B("My Plot"),
                        html.Hr(),
                        dcc.Graph(id="netatmo_data_plot"),
                    ],
                ),
            ],
        ),
    ],
)


@app.callback(
    Output("netatmo_data_plot", "figure"),
    [Input("plot-btn", "n_clicks")],
    [
        State("date-picker-select", "start_date"),
        State("date-picker-select", "end_date"),
        State("variable-select", "value"),
    ],
)
def prepare_animation(n_clicks, start_date, end_date, dataset_var):
    """Prepare the animation to be shown."""
    if n_clicks == 0:
        return domain.get_fig(max_ngrid=0)

    fig_dict, sliders_dict = init_fig_dict(
        domain, dataset_var, frame_duration=300
    )

    # Determine map boundaries and max/min plotted values
    minval_dataset_var = float("inf")
    maxval_dataset_var = -float("inf")

    for fmt in ["%Y-%m-%dT%H:%M:%S", "%Y-%m-%d"]:
        try:
            start_date = dt.strptime(start_date, fmt).replace(hour=0)
            break
        except ValueError:
            pass
    for fmt in ["%Y-%m-%dT%H:%M:%S", "%Y-%m-%d"]:
        try:
            end_date = dt.strptime(end_date, fmt).replace(hour=21)
            break
        except ValueError:
            pass

    for idtg, dtg in enumerate(pd.date_range(start_date, end_date, freq="3H")):
        logger.info("Reading data for %s", dtg)
        df = read_netatmo_data_for_dtg(
            dtg, rootdir=config.general.data_rootdir
        )
        df, _ = remove_irregular_stations(df)
        logger.debug("    * Done. Now adding frame.")

        # Control of map boundaries
        minval_dataset_var = min(minval_dataset_var, min(df[dataset_var]))
        maxval_dataset_var = max(maxval_dataset_var, max(df[dataset_var]))

        frame_dict = fig_dict.copy()
        # plotly.graph_objs.Frame objects have no "frames" property
        frame_dict.pop("frames", None)
        frame_dict["name"] = str(dtg)
        frame, slider_step = generate_single_frame(
            df, dataset_var, frame_duration=30, frame=frame_dict
        )

        if idtg == 0:
            fig_dict["data"] = [frame["data"]] + fig_dict["data"]
        fig_dict["frames"].append(frame)
        sliders_dict["steps"].append(slider_step)

    # Adjust colourbar
    cbar_minnval = 0
    cbar_maxval = maxval_dataset_var
    if minval_dataset_var * maxval_dataset_var < 0:
        cbar_maxval = max(np.abs((minval_dataset_var, maxval_dataset_var)))
        cbar_minnval = -cbar_maxval
    for iframe, frame in enumerate(fig_dict["frames"]):
        frame["data"]["marker"]["cmin"] = cbar_minnval
        frame["data"]["marker"]["cmax"] = cbar_maxval
        fig_dict["frames"][iframe] = frame

    fig_dict["layout"]["sliders"] = [sliders_dict]

    fig = go.Figure(fig_dict)

    return fig


# Run the server
if __name__ == "__main__":
    app.run_server(debug=True)



#!/usr/bin/env python3
"""Common definitions."""
try:
    # From python3.8
    from importlib.metadata import version
except ModuleNotFoundError:
    # Prior to python3.8
    from importlib_metadata import version

try:
    __version__ = version(__name__)
except ModuleNotFoundError:
    __version__ = "?"



#!/usr/bin/env python3
"""Code related to producing plots."""
from pathlib import Path

import humanize
import numpy as np
import pandas as pd
import plotly.express as px
import plotly.graph_objects as go

from .logs import get_logger

# This "config" is plotly's fig.show config, not our parsed config file.
# See <https://github.com/plotly/plotly.js/blob/master/src/plot_api/
#      plot_config.js>
DEF_FIGSHOW_CONFIG = dict(editable=True, displaylogo=False)


def get_obs_scattergeo_trace(df, trace_name=None, marker=None, visible=True):
    """Get a go.Scattergeo object from observations in dataframe df."""
    base_marker = dict(
        color="blue", size=8, opacity=0.75, line=dict(width=1.0)
    )
    if marker is None:
        marker = base_marker
    else:
        new_marker = base_marker.copy()
        new_marker.update(marker)
        marker = new_marker

    # Format popup hover text
    if trace_name is None:
        hovertemplate = []
    else:
        hovertemplate = ["<b>%s</b><br>" % (trace_name)]
    hoverinfo_cols = [c for c in df.columns if not c.startswith("_")]
    for icol, col in enumerate(hoverinfo_cols):
        fmt = None
        if pd.api.types.is_float_dtype(df[col]):
            fmt = ": .5f"
        elif pd.api.types.is_datetime64_any_dtype(df[col]):
            fmt = "| %Y-%m-%d %H:%M:%S"
        if fmt is None:
            hovertemplate.append("%s: %%{customdata[%d]}" % (col, icol))
        else:
            hovertemplate.append("%s: %%{customdata[%d]%s}" % (col, icol, fmt))
    hovertemplate = "<br>".join(hovertemplate)

    return go.Scattergeo(
        name=trace_name,
        lat=df.lat,
        lon=df.lon,
        mode="markers",
        marker=marker,
        customdata=df,
        hovertemplate=hovertemplate,
        hoverlabel=dict(namelength=0),
        visible=visible,
    )


def draw_boundaries(
    fig,
    domain,
    name="boundaries",
    corners=None,
    showlegend=True,
    legendgroup=None,
    **kwargs
):
    """Add to fig line segments connecting the given corners within the domain.

    Args:
        fig (go.Scattergeo): Figure where the boundaries are to be drawn.
        domain (netatmoqc.domains.Domain): Model domain.
        name (str): Name of the boundaries (shown in legend).
            (Default value = "boundaries").
        corners: (x, y) coords of he start and end of the corners of the closed
            boundary to be drawn. If corners is not passed, then the domain
            corners will be used. (Default value = None)
        showlegend (bool): (Default value = True)
        legendgroup (str): Passed to go.Scattergeo. (Default value = None).
        **kwargs: Passed to the "lines" opt of go.Scattergeo.

    Returns:
        go.Scattergeo: Input figure with the passed boundaries drawn.

    """
    if corners is None:
        corners = domain.grid.corners

    # Construct line segments
    # We interpolate using (x, y) instead of (lon, lat) to prevent distorted
    # line segments (segments that do not conform to the used projection).
    segments = []
    npts_per_segment = max(max(domain.grid.nlon, domain.grid.nlat) // 100, 5)
    for istart, start in enumerate(corners):
        end = corners[(istart + 1) % len(corners)]
        segments.append(
            dict(
                x=np.linspace(start[0], end[0], npts_per_segment),
                y=np.linspace(start[1], end[1], npts_per_segment),
            )
        )

    xvals = np.empty(3 * len(segments) * npts_per_segment)
    yvals = np.empty(3 * len(segments) * npts_per_segment)

    # Add all segments as a single trace
    # Put start points in go.Scattergeo's "single-trace" style
    xvals[::3] = np.concatenate([s["x"] for s in segments])
    yvals[::3] = np.concatenate([s["y"] for s in segments])
    lons, lats = domain.proj.xy2lonlat(xvals, yvals)
    # Same for end points
    lats[1::3] = np.roll(lats[::3], 1)
    lons[1::3] = np.roll(lons[::3], 1)
    # Indicate separation between traces
    lats[2::3] = None
    lons[2::3] = None

    fig.add_trace(
        go.Scattergeo(
            name=name,
            lat=lats,
            lon=lons,
            mode="lines",
            line=kwargs,
            legendgroup=legendgroup,
            showlegend=showlegend,
        )
    )

    return fig


def draw_grid_pts(
    fig, grid, display_grid_max_gsize=None, name="Grid", **marker_opts
):
    """Add to fig a trace containing a selection of grid points."""
    if display_grid_max_gsize is None:
        display_grid_max_gsize = grid.x_spacing
    if display_grid_max_gsize <= 0:
        return fig

    grid_draw_every = max(1, int(display_grid_max_gsize / grid.x_spacing))
    lons, lats = grid.ij2lonlat_map()
    if grid_draw_every > 1:
        name += " (every %s point of)" % (humanize.ordinal(grid_draw_every))
    fig.add_trace(
        go.Scattergeo(
            name=name,
            lat=lats[::grid_draw_every, ::grid_draw_every].flatten(),
            lon=lons[::grid_draw_every, ::grid_draw_every].flatten(),
            mode="markers",
            marker=marker_opts,
        )
    )
    return fig


def get_domain_fig(
    domain,
    # Show 1 grid point every "display_grid_max_gsize" meters.
    # Don't show any grid point if display_grid_max_gsize <=0.
    display_grid_max_gsize=0,
    lonrange=None,
    latrange=None,
    ezone=None,
    obs_df=None,
    **kwargs
):
    """Return map representation of "domain" (a ".domains.Domain" object)."""
    # Should we plot the extension zone?
    if ezone is None:
        ezone = domain.grid.nx_ezone > 0

    # Define longitude range
    if lonrange is None:
        if domain.grid.nlon == 1:
            minlon = -180
            maxlon = 180
        elif ezone:
            minlon = domain.ezone_minlon
            maxlon = domain.ezone_maxlon
        else:
            minlon = domain.minlon
            maxlon = domain.maxlon
        lonrange = np.clip((minlon - 2.0, maxlon + 2.0), -180, 180)

    # Define latitude range
    if latrange is None:
        if domain.grid.nlat == 1:
            minlat = -90
            maxlat = 90
        elif ezone:
            minlat = domain.ezone_minlat
            maxlat = domain.ezone_maxlat
        else:
            minlat = domain.minlat
            maxlat = domain.maxlat
        latrange = np.clip((minlat - 2.0, maxlat + 2.0), -90, 90)

    # Make main, base figure
    fig = go.Figure(go.Scattergeo())
    proj2plotly_name = dict(
        merc="mercator", stere="stereographic", lcc="conic conformal"
    )
    fig_center = domain.center_lonlat
    fig.update_geos(
        # resolution is either 50 (higher) or 110 (lower)
        resolution=50,
        projection_type=proj2plotly_name[domain.proj.name],
        center=dict(lon=fig_center[0], lat=fig_center[1]),
        showland=True,
        showcountries=True,
    )
    fig.update_layout(
        height=800,
        margin=dict(r=0, l=0, b=0),
        title=dict(
            text="{} Domain Boundaries and Grid ({} Projection)".format(
                domain.name, domain.proj.full_name
            ),
            xanchor="center",
            x=0.5,
            yanchor="top",
            y=0.9,
        ),
        legend=dict(
            orientation="v",
            xanchor="left",
            x=1.0,
            yanchor="middle",
            y=0.5,
            # Use "reversed" instead of "normal" because we can't directly
            # conrol the order of the layers drawn next. We then draw them
            # from lower to upper and reverse the legend.
            traceorder="reversed",
        ),
        geo=dict(
            lataxis=dict(range=latrange, showgrid=True, dtick=10,),
            lonaxis=dict(range=lonrange, showgrid=True, dtick=15,),
        ),
    )

    # Draw selection of grid points
    # (a) Support grid (if used)
    if domain.thinning_grid is not None:
        fig = draw_grid_pts(
            fig,
            domain.thinning_grid,
            display_grid_max_gsize=display_grid_max_gsize,
            name="Support Grid",
            color="red",
            size=1.5,
            opacity=0.5,
        )
    # (b) Main grid
    fig = draw_grid_pts(
        fig,
        domain.grid,
        display_grid_max_gsize=display_grid_max_gsize,
        color="black",
        size=1.5,
    )

    # Add line segments to mark domain, subdomain and extension zone boundaries
    # (a) Observations
    if obs_df is not None:
        trace = get_obs_scattergeo_trace(obs_df, trace_name="Observations")
        fig.add_trace(trace)

    # (b) Extension zone
    if ezone:
        fig = draw_boundaries(
            fig,
            domain,
            name="Extension Zone",
            corners=domain.grid.ezone_corners,
            color="blue",
            dash="dash",
        )

    # (c) Subdomain boundaries
    if domain.n_subdomains > 1:
        splits = domain.split()
        for isub, subdomain in zip(range(len(splits))[::-1], splits[::-1]):
            showlegend = isub == 0
            sub_name = "Subdomains"
            fig = draw_boundaries(
                fig,
                subdomain,
                name=sub_name,
                showlegend=showlegend,
                legendgroup="Subdomains",
                color="black",
                width=1,
            )

    # (d) Draw main domain doundaries
    fig = draw_boundaries(fig, domain, name="Domain", color="red")

    return fig


def make_clustering_fig(df, domain, **kwargs):
    """Make fig produced by/in "clustering" command/app.

    Args:
        df (pandas.dataframe): Dataframe with clustering data.
        domain (netatmoqc.domains.Domain): Model domain.
        **kwargs: Passed on to the domain.get_fig method.

    Returns:
        go.Scattergeo: Figure depicting the clustering results..

    """
    # Make sure df index is sequential
    df = df.reset_index(drop=True)

    if "cluster_label" not in df.columns:
        df["cluster_label"] = np.nan
    if "original_cluster_label" not in df.columns:
        df["original_cluster_label"] = df["cluster_label"]

    # Add a col to df with descriptive labels for plot
    label_counts = df["cluster_label"].value_counts()
    orig_label_counts = df["original_cluster_label"].value_counts()

    # Vectorize setting of plot labels. About 25x faster than looping over df.
    def _set_plot_label(label, orig_label):
        if label == -1:
            # Obs removed in the normal way with clustering
            legend_label = "Rejected: "
            label_count = label_counts[label]
        elif label == -2:
            # Outliers found after main clustering
            legend_label = "Cluster {}, removed outliers:".format(orig_label)
            label_count = (
                orig_label_counts[orig_label] - label_counts[orig_label]
            )
        elif label == -3:
            legend_label = "Outliers, preliminary clustering:"
            label_count = label_counts[label]
        elif label == -4:
            legend_label = "Missed (due to domain splitting):"
            label_count = label_counts[label]
        elif label == -5:
            legend_label = "Moving stations:"
            label_count = label_counts[label]
        else:
            legend_label = "Cluster {}, accepted:".format(int(label))
            label_count = label_counts[label]
        legend_label += " {} obs".format(int(label_count))
        return legend_label

    set_plot_label = np.vectorize(_set_plot_label, otypes=[str])
    df["_plot_label"] = set_plot_label(
        df["cluster_label"], df["original_cluster_label"]
    )

    # Define colours to be used in the plot
    # See colorscales at <https://plotly.com/python/builtin-colorscales>
    color_discrete_sequence = [
        px.colors.qualitative.Light24,
        px.colors.qualitative.Alphabet,
        px.colors.qualitative.Dark24,
    ][0]

    # Define symbols to be used in the plot
    # See all symbols at
    # <https://plotly.com/python/marker-style/#custom-marker-symbols>
    symbols = [
        "circle",
        "octagon",
        "hexagon2",
        "circle-dot",
        "octagon-dot",
        "circle-cross",
        "circle-x",
    ]

    # Now assign appropriate colors and symbols to each observation row
    unique_labels = df["cluster_label"].unique()
    label2color_map = {}
    label2symbol_map = {}
    for ilab, label in enumerate(unique_labels[unique_labels > -1]):
        color = color_discrete_sequence[ilab % len(color_discrete_sequence)]
        symbol = symbols[(ilab // len(color_discrete_sequence)) % len(symbols)]
        label2color_map[label] = color
        label2symbol_map[label] = symbol

    def _label2symbol(label):
        if label < 0:
            return "x-open"
        return label2symbol_map[label]

    label2symbol = np.vectorize(_label2symbol, otypes=[str])

    def _label2color(label):
        if label < 0:
            return "black"
        return label2color_map[label]

    label2color = np.vectorize(_label2color, otypes=[str])

    df["_plot_color"] = label2color(df["original_cluster_label"])
    df["_plot_symbol"] = label2symbol(df["cluster_label"])

    #########################
    # Now create the figure #
    #########################

    fig = domain.get_fig(max_ngrid=0, **kwargs)
    fig.update_layout(
        height=600,
        clickmode="event+select",
        title=None,
        margin=dict(r=0, l=0, b=0, t=0),
        legend_title="<b>Clustering of Observations</b>",
        legend_traceorder="normal",
    )

    # Add each group as a different trace
    for label, cluster_df in df.groupby("_plot_label", sort=False):

        marker = dict(
            color=cluster_df["_plot_color"].iloc[0],
            symbol=cluster_df["_plot_symbol"].iloc[0],
        )

        if cluster_df["cluster_label"].iloc[0] < 0:
            trace_visible = "legendonly"
            marker.update(dict(size=5, opacity=0.25))
        else:
            trace_visible = True

        trace = get_obs_scattergeo_trace(
            cluster_df, trace_name=label, marker=marker, visible=trace_visible,
        )
        fig.add_trace(trace)

    return fig


def show_cmd_get_fig_from_dataframes(args, dataframes, domain):
    """Make fig produced by "show" command.

    Args:
        args (argparse.Namespace): Parsed command line arguments.
        dataframes (list): List of pandas.dataframe objects with data to be
            plotted as scattergeo.
        domain (netatmoqc.domains.Domain): Model domain.

    Returns:
        go.Scattergeo: Scattergeo fig composed with data in the
        input dataframes.

    """
    logger = get_logger(__name__, args.loglevel)

    fig = domain.get_fig(max_ngrid=0)
    fig.update_layout(
        title=dict(
            text="NetatmoQC Output",
            xanchor="center",
            x=0.5,
            yanchor="top",
            y=0.9,
        ),
        margin=dict(r=0, l=0, b=0),
    )

    # We'll use a fig dict so we can control the order in which data is added
    fig_dict = fig.to_dict()

    new_fig_data = []
    # Add contents from each file as a different trace
    for fname, df in dataframes.items():
        logger.info("Add data from file '%s' (%d obs)", fname, len(df.index))
        if "accepted" in fname:
            trace_visible = True
            marker = dict(color="blue")
        else:
            trace_visible = "legendonly"
            marker = dict(size=5, opacity=0.25, symbol="x-open")
            if "moving" in fname:
                marker["color"] = "red"
            else:
                marker["color"] = "black"

        trace_name = Path(fname).stem.replace("_", " ").title()
        trace_name += " (%d obs)" % (len(df.index))
        trace = get_obs_scattergeo_trace(
            df, trace_name=trace_name, marker=marker, visible=trace_visible
        )
        new_fig_data.append(trace)

    # Add data in reverse order. Needed for legend/layer order.
    fig_dict["data"] = new_fig_data[::-1] + fig_dict["data"]
    return go.Figure(fig_dict)


###########################################################
# Routines only used in apps/scattergeo_timeseries/app.py #
###########################################################


# For animations, see:
# <https://plotly.com/python/animations/#animated-figures-with-graph-objects>
def generate_single_frame(df, dataset_var, frame_duration, frame=None):
    """Generate a single frame of the animation produced by scattergeo app."""
    if frame is None:
        # Dict is mutable and func args are created at func def time
        frame = {}

    # Controlling colour schemes
    if dataset_var in ["temperature"]:
        color_scale = px.colors.diverging.RdBu_r
    else:
        color_scale = px.colors.sequential.haline_r

    marker = dict(
        color=df[dataset_var],
        colorscale=color_scale,
        opacity=0.5,
        line=dict(color="black", width=0.25),
        colorbar=dict(
            titleside="right", ticks="outside", showticksuffix="last",
        ),
    )
    trace = get_obs_scattergeo_trace(df, marker=marker)
    frame["data"] = trace

    slider_step = dict(
        args=[
            [frame["name"]],
            dict(
                frame=dict(duration=frame_duration, redraw=True),
                mode="immediate",
                transition=dict(duration=frame_duration),
            ),
        ],
        label=frame["name"],
        method="animate",
    )

    return frame, slider_step


def init_fig_dict(domain, dataset_var, frame_duration):
    """Initiate dict used to create plotly fig used in scattergeo app."""
    # We make the figure by constructing a dictionary and passing it to the
    # final plotly method.
    fig = domain.get_fig(max_ngrid=0)
    fig.update_geos(resolution=110)
    fig_dict = fig.to_dict()
    # resolution is either 50 (higher) or 110 (lower)
    fig_dict["frames"] = []

    # Figure layout
    fig_dict["layout"]["title"]["text"] = "NetAtmo Data: {}".format(
        dataset_var
    )

    fig_dict["layout"]["updatemenus"] = [
        # <https://plotly.com/python/reference/layout/updatemenus>
        dict(
            type="buttons",
            direction="left",
            showactive=False,
            xanchor="right",
            yanchor="top",
            x=0.1,
            y=0,
            pad=dict(r=10, t=87),
            buttons=[
                dict(
                    label="Play",
                    method="animate",
                    args=[
                        None,
                        dict(
                            frame=dict(duration=frame_duration, redraw=True,),
                            fromcurrent=True,
                            transition=dict(
                                duration=frame_duration / 2,
                                easing="cubic-in-out",
                            ),
                        ),
                    ],
                ),
                dict(
                    label="Pause",
                    method="animate",
                    args=[
                        None,
                        dict(
                            frame=dict(duration=0, redraw=True),
                            mode="immediate",
                            transition=dict(duration=0),
                        ),
                    ],
                ),
            ],
        )
    ]

    sliders_dict = dict(
        active=0,
        len=0.9,
        yanchor="top",
        xanchor="left",
        x=0.1,
        y=0,
        pad=dict(b=10, t=50),
        currentvalue=dict(
            font=dict(size=20), prefix="DTG: ", visible=True, xanchor="right",
        ),
        transition=dict(duration=frame_duration / 2, easing="cubic-in-out",),
        steps=[],
    )

    return fig_dict, sliders_dict



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['netatmoqc',
 'netatmoqc.apps',
 'netatmoqc.apps.clustering',
 'netatmoqc.apps.scattergeo_timeseries']

package_data = \
{'': ['*'],
 'netatmoqc': ['bin/*'],
 'netatmoqc.apps.clustering': ['assets/*'],
 'netatmoqc.apps.scattergeo_timeseries': ['assets/*']}

install_requires = \
['attrs>=20.2.0,<21.0.0',
 'dash-daq>=0.5.0,<0.6.0',
 'dash>=1.13.4,<2.0.0',
 'dotmap>=1.3.17,<2.0.0',
 'flask-caching>=1.9.0,<2.0.0',
 'hdbscan>=0.8.26,<0.9.0',
 'humanize>=2.6.0,<3.0.0',
 'importlib-metadata>=1.7.0,<2.0.0',
 'joblib>=0.15.1,<0.16.0',
 'llvmlite==0.33',
 'numba>=0.50.1,<0.51.0',
 'numpy>=1.19.0,<2.0.0',
 'pandas>=1.0.5,<2.0.0',
 'plotly>=4.8.2,<5.0.0',
 'psutil>=5.7.0,<6.0.0',
 'pyproj>=2.6.1,<3.0.0',
 'pytz>=2020.1,<2021.0',
 'redis>=3.5.3,<4.0.0',
 'scikit-learn>=0.23.1,<0.24.0',
 'toml>=0.10.1,<0.11.0']

extras_require = \
{':sys_platform != "darwin"': ['tbb>=2020.0.133,<2021.0.0'],
 'mpi': ['mpi4py>=3.0.3,<4.0.0']}

entry_points = \
{'console_scripts': ['netatmoqc = netatmoqc.main:main']}

setup_kwargs = {
    'name': 'netatmoqc',
    'version': '0.3.4',
    'description': 'Use machine learning clustering methods to perform quality control over NetAtmo data',
    'long_description': '# NetAtmoQC\n\n\n**Table of Contents**\n\n[[_TOC_]]\n\n\n### About\n\n`netatmoqc` is a python package that uses Machine Learning Clustering methods to\nquality-control observations collected from [NetAtmo](https://www.netatmo.com/en-gb)\nweather stations. It has so far been developed at [SMHI](https://www.smhi.se/en)\nas part of the [iObs](https://wiki.neic.no/wiki/IOBS) project.\n\nPlease note that this package is still in its development/implementation stage.\nAs such, it may (certainly does) contain (hopefully minor) bugs and lack on\ndocumentation. If you wish to collaborate, suggest features or report issues,\nplease contact [Paulo Medeiros (SMHI)](mailto:paulo.medeiros@smhi.se).\n\nA note about this file: We have used the\n[Markdown format](https://docs.gitlab.com/ee/user/markdown.html) throughout.\nYou should, however, be able to read it reasonably well with your plain text\nprocessor of choice. Please disregard the formatting marks in this case.\n\nSee also the [project\'s Wiki](https://source.coderefinery.org/iOBS/wp2/task-2-3/netatmoqc/-/wikis/home) for more information.\n\n### System Requirements\n\n* python >=3.6.10\n* A C compiler\n\n* **Optional**: Ability to compile and run MPI applications.\n\n    The system needs to have a working installation of an MPI library. Having\n    [Open MPI](https://www.open-mpi.org/) should be fine, but there are other\n    options.\n\n    This requirement is usually already fulfilled in HPC facilities, although,\n    in some cases, you might need to load a module (e.g., `module load openmpi`).\n    Please check with your HPC support if you have doubts about this.\n\n    **NB.:** If this requirement is not fulfilled, you won\'t be able to run\n    `netatmoqc` using MPI, even if you manage to follow the MPI-related\n    installation instructions presented later on in this file.\n\n    If you don\'t have a working MPI library installed in your system but use,\n    for instance, [conda](https://docs.conda.io/projects/conda/en/latest/glossary.html#anaconda-glossary)\n    to manage your environments/source packages, then running the following\n    commands should get it working:\n\n        conda install -c conda-forge openmpi\n        conda install gxx_linux-64\n\n* **Only for\n[Developer-Mode Installtion](#developer-mode-installation):**\n\n    * [`poetry`](https://python-poetry.org), which can be installed by running\n\n            curl -sSL https://raw.githubusercontent.com/python-poetry/poetry/master/get-poetry.py | python3\n\n\n### Installation\n\nBefore proceeding, please make sure that your system fulfils the appropriate\n[system requirements](#system-requirements). If you plan to just use the code\nwithout modifying it, please follow one of the installation methods presented\nin the [Regular Installation section](#regular-installation). However, if you\nneed/wish to modify the code in any way, then please proceed as indicated in the\n[Developer-Mode Installtion section](#developer-mode-installation).\n\n**N.B.:** In any case, you will be presented with instructions for installation\nwith and without MPI support. *You only need to follow one such set of\ninstructions*. Mind, however, the extra [system requirements](#system-requirements)\nthat apply if you choose to install `netatmoqc` with MPI support.\n\n\n#### Regular Installation\n##### Regular Installation from PyPi\n:point_right: Easiest method if you just want to use the code and don\'t want to\nlook at the source code at all.\n\n* Install *without* MPI support:\n\n        pip install netatmoqc\n\n* Install with MPI support:\n\n        pip install netatmoqc[mpi]\n\n\n##### Regular Installation Directly From The Git Repo\n\n:point_right: Similar to a [regular installation from PyPi](#regular-installation-from-pypi),\nbut retrieves the code from the git repo instead (which is usually updated more\noften).\n\n* Install *without* MPI support:\n\n        pip install "git+https://source.coderefinery.org/iOBS/wp2/task-2-3/netatmoqc"\n\n* Install with MPI support:\n\n        pip install "netatmoqc[mpi] @ git+https://source.coderefinery.org/iOBS/wp2/task-2-3/netatmoqc"\n\n\n##### Regular Installation From Downloaded Source\n\n:point_right: For those who have `netatmoqc`\'s source code in a local directory,\nwish to install it from there, but also don\'t want to modify any code.\n\n* Install *without* MPI support:\n\n        pip install .\n\n* Install with MPI support:\n\n        pip install ".[mpi]"\n\n#### Developer Mode Installation\n\n:point_right: For those who need/wish to make changes to `netatmoqc`\'s\nsource code, or use code from a different branch than `master`.\n\n* Install *without* MPI support:\n\n        poetry install\n\n* Install with MPI support:\n\n        poetry install --extras "mpi"\n\nInstalling in "developer mode" means that changes made in any of the package\'s\nsource files become visible as soon as the package is reloaded.\n\n:wrench: **Recommendation to contributors:** Before making your first commit to\nthe repo, please also run the following:\n\n    pre-commit install\n\nThis sets up the git hook scripts defined in the\n[.pre-commit-config.yaml](.pre-commit-config.yaml) file and only needs to be run\n(i) before the first commit, and (ii) after having modified the\n[.pre-commit-config.yaml](.pre-commit-config.yaml) file. The\n[pre-commit](https://pre-commit.com) package is installed when you run any of\nthe `poetry install` commands listed above.\n\n\n### After Installation: Configuration File\n\nAfter successful installation, a `netatmoqc` command will become available in\nyour environment. However, before you can use `netatmoqc` (apart from the `-h`\noption), you will need a configuration file written in the\n[TOML](https://en.wikipedia.org/wiki/TOML) format.\n\nPlease take a look at the\n[docs/minimal_config_example.toml](docs/minimal_config_example.toml) and\n[docs/more_complete_config_example.toml](docs/more_complete_config_example.toml)\nfiles, as well as the [project\'s Wiki](https://source.coderefinery.org/iOBS/wp2/task-2-3/netatmoqc/-/wikis/home), for more information about the configuration file.\n\n\n`netatmoqc` assumes that one of the following (whichever is first encountered)\nis your configuration file :\n\n1. A *full file path* specified via the `NETATMOQC_CONFIG_PATH` envvar\n2. A `config.toml` file located in the directory where `netatmoqc` is called\n3. `$HOME/.netatmoqc/config.toml`\n\n\n### Usage\nAfter completing the setup, you should be able to run\n\n    netatmoqc [opts] SUBCOMMAND [subcommand_opts]\n\nwhere `[opts]` and `[subcommand_opts]` denote optional command line arguments\nthat apply, respectively, to `netatmoqc` in general and to `SUBCOMMAND`\nspecifically.\n\n**Please run `netatmoqc -h` for information** about the supported subcommands\nand general `netatmoqc` options. For info about specific subcommands and the\noptions that apply to them only, **please run `netatmoqc SUBCOMMAND -h`** (note\nthat the `-h` goes after the subcommand in this case).\n\n**N.B.:** When using the (preferred and default) clustering method\n[HDBSCAN](https://hdbscan.readthedocs.io/en/latest/index.html), a typical\n`netatmoqc` run (~40000 unique station IDs per DTG) needs ca. **6 GB** RAM and\ntakes between 1 and 2 minutes to complete (on a machine equipped with Intel(R)\nXeon(R) CPU E5-2640 v3 @ 2.60GHz, 16 physical cores). Other implemented\nclustering strategies have more modest RAM requirements, but:\n  * [DBSCAN](https://scikit-learn.org/stable/modules/generated/sklearn.cluster.DBSCAN.html)\n  results are not as good as HDBSCAN\'s in our context\n  * [OPTICS](https://scikit-learn.org/stable/modules/generated/sklearn.cluster.OPTICS.html)\n  produces similar results as HDBSCAN but runs *much* slower\n\n\n#### Parallelism (single-host or MPI)\n\nThe `select` subcommand supports parallelism over DTGs. How to activate it\ndepends on whether you wish to run `netatmoqc` on a single host or if you wish\nto distribute computations over different computers (e.g. on an HPC cluster).\n\n  * If you are running `netatmoqc` in a single host, then you can export the\n    environment variable `NETATMOQC_MAX_PYTHON_PROCS` to any value larger\n    than 0 and run the code as usual.\n\n  * If you wish to run `netatmoqc` with MPI, then you must have installed it\n    with MPI support. Assuming this is the case, you can then run the code as\n\n        mpiexec -n 1 [-usize N] netatmoqc --mpi [opts] select [subcommand_opts]\n\n    Notice that:\n    * Arguments between square brackets are optional\n    * The `--mpi` switch must come before any subcommand\n    * **The value "1" in `-n 1` is mandatory.** The code will always start with one\n      "manager" task which will dynamically spawn new worker tasks as needed\n      (up to a maximum number).\n    * If `-usize N` is passed, then `N` should be an integer greater than zero.\n      `N` defines the maximum number of extra workers that the manager task is\n      allowed to spawn if necessary.\n    * If `-usize N` is not passed, then:\n      * If the run is part of a submitted job managed by SLURM or PBS, then `N`\n        will be automatically determined from the options passed to the\n        scheduler (e.g. `--nnodes`, `--ntasks`, `--mem-per-cpu`, etc for SLURM).\n      * If the run is running interactive: `N` will take the value of the\n        environment variable `NETATMOQC_MAX_PYTHON_PROCS` if set, or, otherwise,\n        will be set to 1.\n    * No more than `length(DTGs)` new worker tasks will be spawn\n\nLast, but not least: Please keep in mind the RAM requirements discussed in the\n[usage](#usage) section.\n',
    'author': 'Paulo V. C. Medeiros',
    'author_email': 'paulo.medeiros@smhi.se',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://source.coderefinery.org/iOBS/wp2/task-2-3/netatmoqc',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.6.10,<4.0.0',
}


setup(**setup_kwargs)



#!/usr/bin/env python3
"""Routines related to MPI usage."""
import logging
import os

try:
    from mpi4py import MPI
    from mpi4py.futures import MPIPoolExecutor

    MPI4PY_IMPORT_ERROR = None
except ImportError as err:
    MPI4PY_IMPORT_ERROR = err
from .logs import logcolor

logger = logging.getLogger(__name__)


def mpi_parallel(fun, iterable):
    """Run function "fun" in parallel over "iterable" using MPI."""
    if MPI4PY_IMPORT_ERROR is not None:
        msg = "%s. %s%s%s" % (
            MPI4PY_IMPORT_ERROR,
            logcolor.red,
            "Support to MPI is unavailable!",
            logcolor.reset,
        )
        raise ImportError(msg) from MPI4PY_IMPORT_ERROR

    # Prevent using mpiexec with n>1
    comm = MPI.COMM_WORLD
    size = comm.Get_size()
    rank = comm.Get_rank()
    if size > 1:
        if rank == 0:
            raise ValueError(
                "Received '-n %d' from the MPI runner. Please " % (size)
                + "use '-n 1' when running this application with MPI, "
                + "and then select the maximum number N of parallel "
                + "MPI tasks by passing '-usize N'"
            )

    # Establish max #workers that will be dynamically spawn
    # (i.a) Check if universe size passed explicitly (e.g., via "-usize")
    univ_size = comm.Get_attr(MPI.UNIVERSE_SIZE)
    # (i.b) If not (i.a), check if we can guess from the scheduler's env
    if univ_size is None:
        univ_size = os.getenv("SLURM_NTASKS", os.getenv("PBS_NP", None))
    # (i.c) If not (i.b), assume it's a local run
    if univ_size is None:
        univ_size = int(os.getenv("NETATMOQC_MAX_PYTHON_PROCS", "1"))
    # (ii) Check how many tasks will actually be performed
    max_workers = min(len(iterable), int(univ_size))

    # Spawn workers and perform tasks
    with MPIPoolExecutor(max_workers=max_workers) as executor:
        results = executor.map(fun, iterable)
    return list(results)
